/*
!C******************************************************************

!Description: 
	Functions for writing information to be used in generating
	the MODIS fire Climate Modeling Grid (CMG) product.  This
	ability was added to the Level 2 code to prevent the CMG
	code from having to use both L2 MOD14 and MOD03 granules to
	make the monthly CMG product.  Because each MOD03 granule is
	~4 times larger than each MOD14 granule, this greatly reduces
	the amount of input data needed for CMG production.

!Revision History:
	Original version, April 2002, L. Giglio (giglio@hades.gsfc.nasa.gov).
	Revised June 2006 to separate day and night observations.

!Design Notes

!END*************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include "mfhdf.h"
#include "cmg.h"
#include "error.h"
#include "fire.h"


typedef unsigned short USHORT;



/* 0.25 degree bins covering the entire globe */
#define CMG_BIN_SIZE 0.25F
#define CMG_ROWS 720
#define CMG_COLS 1440
#define CMG_ELEMS (CMG_ROWS * CMG_COLS)



typedef struct {
	USHORT	total, water, missing, cloud;
	USHORT	unknown, fire;
	} GridCell;



/* CMG_VALUES must always be last enumeration */
enum {CMG_X, CMG_Y, CMG_TOTAL, CMG_WATER, CMG_MISSING, CMG_CLOUD,
	CMG_UNKNOWN, CMG_FIRE, CMG_VALUES};


static GridCell *cmg_day = (GridCell *) NULL;
static GridCell *cmg_night = (GridCell *) NULL;


#ifndef ECS
extern int verbose;
#endif



int cmg_init(int flag)
{
	if (!flag) return 0;

	if ( !(cmg_day = calloc(CMG_ELEMS, sizeof(GridCell))) )
		return error(ERR_F_MEM);
	if ( !(cmg_night = calloc(CMG_ELEMS, sizeof(GridCell))) )
		return error(ERR_F_MEM);

	return 0;
}



void cmg_append(float lat, float lon, uint8 maskval, int day)
{
	int row, col;
	long i;
	GridCell *cmg;

	if (day) cmg = cmg_day; else cmg = cmg_night;

	/* exclude missing or bad data */
	if ((lat < -90.0F) || (lat > 90.0F) || (lon < -180.0F) || (lon > 180.0F)) return;
	row = (uint16) ((90.0F - lat) / CMG_BIN_SIZE);
	col = (uint16) ((lon + 180.0F) / CMG_BIN_SIZE);

	if (row >= CMG_ROWS) row = CMG_ROWS - 1;
	if (col >= CMG_COLS) col = CMG_COLS - 1;

	i = row * CMG_COLS + col;

	cmg[i].total++;

	switch (maskval) {
		case CLASS_WATER:
			cmg[i].water++;
			break;

		case CLASS_CLOUD:
			cmg[i].cloud++;
			break;

		case CLASS_NOPROC_MISS:
			cmg[i].missing++;
			break;

		case CLASS_UNKNOWN:
			cmg[i].unknown++;
			break;

		case CLASS_FIRE_LOW:
		case CLASS_FIRE_MED:
		case CLASS_FIRE_HIGH:
			cmg[i].fire++;
			break;
		}
}



void cmg_free(int flag)
{
	if (flag) {
		if (cmg_day) free(cmg_day);
		if (cmg_night) free(cmg_night);

		/* ensure these are null for debugging and to prevent problems
		if cmg_free() is mistakenly called more than once */
		cmg_day = cmg_night = (GridCell *) NULL;
		}
}



/* keep this at or above ~1000 otherwise HDF internal compression will
become inefficient */
#define BUFF_ROWS 2000



int cmg_write(int32 sd_id, int day)
/*
!C*****************************************************************************

!Description:
	Create and writes the SDS's comprising data layers that will
	be used to create the fire CMG product from the L2 fire product.

!Input Parameters:
	sd_id	HDF SD ID.
	day	flag: non-zero = write daytime SDS, otherwise write nighttime

!Output Parameters:
	None.

!Input/Ouput Parameters:
	None.

!Returns:
	Returns 0 if successful, non-zero otherwise.

!Revision History:
	Original version, L. Giglio (April 2002)
	Revised June 2006 to handle separate day and night observations and
	enable internal HDF compression.

!Design Notes

!END
****************************************************************************/
{
	int i, k;
	int32 sds_id;
	int32 dim_sizes[2];
	int32 start[2], edges[2];
	int32 dim_id;

	int ncells;
	GridCell *cmg;
	char *dimname, *sdsname;

	HDF_CHUNK_DEF cdef;
	uint16 buff[BUFF_ROWS][CMG_VALUES];



	if (day) {
		cmg = cmg_day;
		dimname = "cmg_cells_day";
		sdsname = "CMG_day";
		}
	else {
		cmg = cmg_night;
		dimname = "cmg_cells_night";
		sdsname = "CMG_night";
		}


	/* count number of nonzero cells */
	ncells = 0;
	for (i = 0; i < CMG_ELEMS; i++)
		if (cmg[i].total) ncells++;

#ifndef ECS
if (verbose) printf("%d %s cells\n", ncells, sdsname);
#endif


	/* leave if there are no cells to write */
	if (!ncells) return 0;


	dim_sizes[0] = ncells;
	dim_sizes[1] = CMG_VALUES;

	if ((sds_id = SDcreate(sd_id, sdsname, DFNT_UINT16, 2, dim_sizes))
		== FAIL) return error(ERR_HDF);

	/* chunk and compress CMG data only if a substantial number of
	values are to be written */
	if (ncells >= 200) {
		cdef.chunk_lengths[0] = cdef.comp.chunk_lengths[0] = BUFF_ROWS;
		cdef.chunk_lengths[1] = cdef.comp.chunk_lengths[1] = CMG_VALUES;
		cdef.comp.comp_type = COMP_CODE_DEFLATE;
		cdef.comp.cinfo.deflate.level = 8;

		if (SDsetchunk(sds_id, cdef, HDF_CHUNK | HDF_COMP) == FAIL)
			return error(ERR_HDF);
		}

	if ((dim_id = SDgetdimid(sds_id, 0)) == FAIL)
		return error(ERR_HDF);
	if (SDsetdimname(dim_id, dimname) == FAIL)
		return error(ERR_HDF);

	if ((dim_id = SDgetdimid(sds_id, 1)) == FAIL)
		return error(ERR_HDF);
	if (SDsetdimname(dim_id, "cmg_values") == FAIL)
		return error(ERR_HDF);

	if (SDsetdatastrs(sds_id, "Climate Modeling Grid information", NULL, NULL, NULL) == FAIL)
		return error(ERR_HDF);


	start[1] = 0;
	edges[1] = CMG_VALUES;

	start[0] = 0;
	edges[0] = BUFF_ROWS;
	k = 0;
	for (i = 0; i < CMG_ELEMS; i++) {
		if (!cmg[i].total) continue;

		if (k >= BUFF_ROWS) {
			if (SDwritedata(sds_id, start, NULL, edges, buff) == FAIL)
				return error(ERR_HDF);
			k = 0;
			start[0] += BUFF_ROWS;
			}

		buff[k][CMG_X] = i % CMG_COLS;
		buff[k][CMG_Y] = i / CMG_COLS;
		buff[k][CMG_TOTAL] = cmg[i].total;
		buff[k][CMG_WATER] = cmg[i].water;
		buff[k][CMG_MISSING] = cmg[i].missing;
		buff[k][CMG_CLOUD] = cmg[i].cloud;
		buff[k][CMG_UNKNOWN] = cmg[i].unknown;
		buff[k][CMG_FIRE] = cmg[i].fire;
		k++;
		}

	/* write partial buffer */
	if (k) {
		edges[0] = k;
		if (SDwritedata(sds_id, start, NULL, edges, buff) == FAIL)
			return error(ERR_HDF);
		}

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);

	return 0;
}


