#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mfhdf.h"

#include "const.h"
#include "types.h"
#include "tb.h"
#include "fire.h"
#include "detfire.h"
#include "firetab.h"
#include "cloud.h"
#include "bowtie.h"


#ifdef OLD_RED_HAT
#include "old_red_hat.h"
#endif

#ifdef NO_SINGLE_MATH
#include "no_sing_math.h"
#endif 


#undef max
#define max(a,b) (a < b ? b : a)


#define MAX_WIN_SIZE 21
#define MAX_WIN_PIX (MAX_WIN_SIZE * MAX_WIN_SIZE)


typedef struct {
	/* thresholds used to identify potential fire pixels */
	float	thresh_PF_T21;	/* K */
	float	thresh_PF_DT;	/* K */
	float	thresh_PF_R2;	/* near-IR reflectance threshold, daytime algorithm only */

	/* thresholds used to identify valid bkg. pixels */
	float	thresh_bkg_T21;	/* K */
	float	thresh_bkg_DT;	/* K */

	float	thresh_T21;	/* K */

	float	minbkg_DT;	/* K; minimum difference above bkg. DT */

	float	devrp_T31;	/* K; T31 absolute deviation reduction parameter (daytime only) */
	} AlgParams;



static int BkgPixels(int ic, int jc, Scan *PrevScan, Scan *CurrScan, Scan *NextScan,
	float *BuffT21, float *BuffT31, float *BuffDT,
	float *BuffR1, float *BuffR2, float *BuffR7, int *nref,
	float *BkgFireBuffT21, float *BkgFireBuffT31,
	float *BkgFireBuffDT, int *nbfire, int *winparam, int *nwater);

static void GetValidBkgPixels(int ic, int jc, int r,
	Scan *PrevScan, Scan *CurrScan, Scan *NextScan,
	float *BuffT21, float *BuffT31, float *BuffDT, int *nvalid,
	float *BuffR1, float *BuffR2, float *BuffR7, int *nref,
	float *BkgFireBuffT21, float *BkgFireBuffT31,
	float *BkgFireBuffDT, int *nbfire, int *nwater);

static void CountAdjacentPixels(int ic, int jc,
	Scan *PrevScan, Scan *CurrScan, Scan *NextScan,
	int *AdjCloud, int *AdjWater);

static void compute_stats(float *x, int n, float *mean, float *mad);
static int SunGlint(float, float, float, float, float, float, float *);
static int confidence(float T21, float z21, float DT, float, float,
	int AdjCloud, int AdjWater, int day);
static float fire_power(float T21, float MeanT21, int sample);
static void compute_ndvi(float *R1, float *R2, int n, float *NDVI);
static int bkg_water_pix(float *R2, float *R7, float *NDVI, int n);
static float pixel_area(int sample);

static float S(float x, float a, float b);
static float fixang(float);



static AlgParams DayParams = {
	310.0F,	10.0F,	/* no fire */
	0.30F,		/* reflectance */

	325.0F,	20.0F,	/* valid bkg. */

	360.0F,

	6.0F,

	4.0F
	};


static AlgParams NightParams = {
	305.0F,	10.0F,	/* no fire */
	0.0F,		/* dummy "nighttime reflectance" */

	310.0F,	10.0F,	/* valid bkg. */

	320.0F,

	6.0F,

	1000.0F		/* dummy nighttime value that will cause test to always be passed */
	};


/* output buffers for the fire product */
extern uint8	FireMask[LINES_PER_SCAN_1KM][SAMPLES_1KM];
extern uint32	FireQA[LINES_PER_SCAN_1KM][SAMPLES_1KM];

extern UBYTE iprev[SAMPLES_1KM], inext[SAMPLES_1KM];
#ifndef ECS
extern int dump_fires;
#endif



void prescreen(Scan *scan)
/*
!C*****************************************************************************
!Description:
	Pre-screen current buffer prior to performing detection algorithm.

!Input Parameters:
	scan		pointer to current scan buffer

!Output Parameters:
	None.

!Input/Output Parameters:
	None.

!Returns:
	Nothing.

!Revision History

!Team-unique Header:

!References and Credits

!Design Notes

!END***************************************************************************
*/
{
	int i, j;
	int land;
	int day;

	int glint;		/* flag: glint = non-zero */
	float GlintAng;		/* degrees */

	float T21, T22, T31;
	float DT;
	AlgParams *ap;
	float RelAzAng;

	int pftest1, pftest2, pftest3;

	const float MaxT22 = 329.0F;	/* use T21 if T22 above this threshold */

	static PixInfo NullPixInfo = {0,0,0,0,0,0,0,0};


	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		for (j = 0; j < SAMPLES_1KM; j++) {
			/* clear internal flags */
			scan->pixflags[i][j] = NullPixInfo;

			T21 = scan->T21[i][j];
			T22 = scan->T22[i][j];
			T31 = scan->T31[i][j];

			if (T31 < 0.0F) continue;

			/*
			Use band 22 if available and not saturated,
			otherwise use band 21.  Here we'll actually
			replace the band 21 brightness temperature
			with band 22 provided 22 is available and
			not saturated.
			*/
			if (T22 > 0.0F && T22 < MaxT22)
				/* 22 available and unsaturated */
				T21 = scan->T21[i][j] = T22;
			else {
				/* 22 unavailable or saturated */
				if (T21 > 0.0) {
					/* 22 unavailable or saturated, 21 available */
					T22 = scan->T22[i][j] = T21;

					/* set internal flag */
					scan->pixflags[i][j].high_range = 1;
					}
				else {
					/* 22 unavailable or saturated, 21 unavailable */
					if (T22 > 0.0F)
						/* 22 saturated, 21 unavailable */
						T21 = scan->T21[i][j] = T22;
					else
						/* both 21 and 22 unavailable */
						continue;
					}
				}


			/* set day/night flag */
/* FIX: handle missing */
			if (scan->SolZenAng[i][j] < 85.0F) {
				scan->pixflags[i][j].day = 1;
				day = -1;


				/* sun-glint test */
				RelAzAng = fixang(scan->SolAzAng[i][j] - scan->ViewAzAng[i][j]);
				glint = SunGlint(scan->R1[i][j], scan->R2[i][j], scan->R7[i][j],
					scan->SolZenAng[i][j], scan->ViewZenAng[i][j],
					RelAzAng, &GlintAng);

				scan->pixflags[i][j].glint_level = glint;
				}
			else
				day = 0;

			/* treat MOD03 "land" and "ephemeral water" as land */
			land = ((scan->landsea[i][j] == 1) || (scan->landsea[i][j] == 4));

			if (!land) {
				/* set internal flag */
				scan->pixflags[i][j].water = 1;
				continue;
				}

			/* cloud mask */
			if (scan->T32[i][j] > 0.0F) {
				if (iscloud(scan->R1[i][j], scan->R2[i][j],
					scan->T32[i][j])) {
					scan->pixflags[i][j].cloud = 1;
					continue;
					}
				}
			/* else set missing cloud mask flag */

			DT = T21 - T31;

			if (day)
				ap = &DayParams;
			else
				ap = &NightParams;

			pftest1 = T21 > ap->thresh_PF_T21;
			pftest2 = DT > ap->thresh_PF_DT;
			pftest3 = scan->R2[i][j] < ap->thresh_PF_R2;

			/* potential fire test */
			if (pftest1 && pftest2 && pftest3)
				scan->pixflags[i][j].potential_fire = 1;

			/* valid background pixel test */
			if ((T21 > ap->thresh_bkg_T21) && (DT > ap->thresh_bkg_DT))
				scan->pixflags[i][j].bkg_fire = 1;
			else
				scan->pixflags[i][j].valid_bkg_pix = 1;

			scan->pixflags[i][j].clear_land = 1;
			}
		}
}




int detect(Scan *PrevScan, Scan *scan, Scan *NextScan, int first_line,
	FireCounts *counts)
/*
!C*****************************************************************************
!Description:
	Apply detection algorithm to current scan.

!Input Parameters:
	PrevScan,
	CurrScan,
	NextScan	pointers to previous, current, and next scan buffers

	first_line

!Output Parameters:
	counts

!Input/Output Parameters:
	None.

!Returns:
	0 if no errors, non-zero if fire pixel table functions encountered
	error allocating memory.

!Revision History
	Original version April 2001, L. Giglio.


!END***************************************************************************
*/
{
	int i, j;
	float T21, T31;
	float DT;
	int day;		/* flag: day = non-zero, night = 0 */
	AlgParams *ap;

	int test1, test2, test3, test4, test5, test6;
	int fire;		/* flag: fire pixel */

	int winparam;
	int L;			/* background window size (L x L) */

	float MeanT21, MeanDT, MeanT31;
	float MadT21, MadDT, MadT31;
	float z21;
	const float eps = 1.0e-10F;

	int AdjCloud, AdjWater;
	int C;			/* detection confidence */

	float power;

	int st;

	int nvalid;
	float BuffT21[MAX_WIN_PIX], BuffT31[MAX_WIN_PIX], BuffDT[MAX_WIN_PIX];
	int nref;
	float BuffR1[MAX_WIN_PIX], BuffR2[MAX_WIN_PIX], BuffR7[MAX_WIN_PIX];
	float BuffNDVI[MAX_WIN_PIX];

	int nwater_cont;

	float BkgFireMeanT21, BkgFireMeanDT, BkgFireMeanT31;
	float BkgFireMadT21, BkgFireMadDT, BkgFireMadT31;

	float fvalid;
	int nbfire;
	int nvalid_land;
	int nwater;
	float BkgFireBuffT21[MAX_WIN_PIX], BkgFireBuffT31[MAX_WIN_PIX], BkgFireBuffDT[MAX_WIN_PIX];



	/* initialize the "inext" and "iprev" arrays for each scan */
	init_bowtie_tables(PrevScan, scan, NextScan);


	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		for (j = 0; j < SAMPLES_1KM; j++) {
			/* clear all QA bits */
			FireQA[i][j] = 0;

			T21 = scan->T21[i][j];
			T31 = scan->T31[i][j];
			DT = T21 - T31;
			day = scan->pixflags[i][j].day;


			/* this test was moved up here in v5.0.1 to correct a
			long standing bug in the counting of day and night pixels */
			if (day) {
				ap = &DayParams;

				/* set day/night algorithm flag in product */
				FireQA[i][j] |= 0x10;

				counts->NumDayPix++;
				}
			else {
				ap = &NightParams;

				counts->NumNightPix++;
				}

			/* missing data? */
			if ((T21 < 0.0F) || (T31 < 0.0F)) {
				FireMask[i][j] = CLASS_NOPROC_MISS;
				FireQA[i][j] |= 3;	/* MODLAND QA */

				counts->NumMissingPix++;
				continue;
				}


			if ( !scan->pixflags[i][j].high_range )
				FireQA[i][j] |= 4;

			/* update counts of glint pixels and set glint bit */
			if (scan->pixflags[i][j].glint_level) {
				counts->NumGlintPix++;
				FireQA[i][j] |= (scan->pixflags[i][j].glint_level << 23);
				}


			/* water? */
			if (scan->pixflags[i][j].water) {
				FireMask[i][j] = CLASS_WATER;
				FireQA[i][j] |= 3;	/* MODLAND QA */

				counts->NumWaterPix++;
				if (scan->pixflags[i][j].cloud)
					counts->NumWaterCloudPix++;
				continue;
				}

			/* count non-water pixels containing valid data as land */
			counts->NumLandPix++;

			/* cloud? */
			if (scan->pixflags[i][j].cloud) {
				FireMask[i][j] = CLASS_CLOUD;
				FireQA[i][j] |= 2;	/* MODLAND QA */

				counts->NumLandCloudPix++;
				continue;
				}


			if ( !scan->pixflags[i][j].potential_fire ) {
				FireMask[i][j] = CLASS_NO_FIRE;
				counts->NumNoFirePix++;
				continue;
				}
			else
				/* set algorithm QA: potential fire pixel */
				FireQA[i][j] |= 0x20;


			/* compute background statistics for all potential fire pixels */
			nvalid = BkgPixels(i, j, PrevScan, scan, NextScan,
				BuffT21, BuffT31, BuffDT,
				BuffR1, BuffR2, BuffR7, &nref,
				BkgFireBuffT21, BkgFireBuffT31, BkgFireBuffDT,
				&nbfire, &winparam, &nwater);

			nvalid_land = nvalid + nbfire;

L = 2*winparam + 1;

			if (nref) {
				compute_ndvi(BuffR1, BuffR2, nref, BuffNDVI);
				nwater_cont = bkg_water_pix(BuffR2, BuffR7, BuffNDVI, nref);
				}
			else
				nwater_cont = 0;


			if (nbfire > 0) {
				compute_stats(BkgFireBuffT21, nbfire,
					&BkgFireMeanT21, &BkgFireMadT21);
				compute_stats(BkgFireBuffDT, nbfire,
					&BkgFireMeanDT, &BkgFireMadDT);
				compute_stats(BkgFireBuffT31, nbfire,
					&BkgFireMeanT31, &BkgFireMadT31);
				}
			else {
				BkgFireMeanT21 = BkgFireMeanT31 = BkgFireMeanDT = 0.0F;
				BkgFireMadT21 = BkgFireMadT31 = BkgFireMadDT = 0.0F;
				}


			if (nvalid > 0) {
				compute_stats(BuffT21, nvalid, &MeanT21, &MadT21);
				compute_stats(BuffDT, nvalid, &MeanDT, &MadDT);
				compute_stats(BuffT31, nvalid, &MeanT31, &MadT31);

				fvalid = (float) nvalid / (float) nvalid_land;

				z21 = (T21 - MeanT21) / (MadT21 + eps);

				test2 = DT > (MeanDT + 3.5F * MadDT);
				test3 = DT > (MeanDT + ap->minbkg_DT);
				test4 = T21 > (MeanT21 + 3.0F * MadT21);
				test5 = T31 > (MeanT31 + MadT31 - ap->devrp_T31);
				test6 = BkgFireMadT21 > 5.0F;
				}
			else {
				MeanT21 = MeanT31 = MeanDT = 0.0F;
				MadT21 = MadT31 = MadDT = 0.0F;
				z21 = 0.0F;
				test2 = test3 = test4 = test5 = test6 = 0;
				fvalid = 0.0F;
				}

			test1 = T21 > ap->thresh_T21;


			/* set background characterization field */
			FireQA[i][j] |= (winparam << 7);

			/* save individual test results */
			if (test1) FireQA[i][j] |= 0x800;
			if (test2) FireQA[i][j] |= 0x1000;
			if (test3) FireQA[i][j] |= 0x2000;
			if (test4) FireQA[i][j] |= 0x4000;
			if (test5) FireQA[i][j] |= 0x8000;
			if (test6) FireQA[i][j] |= 0x10000;

			fire = test1 || (test2 && test3 && test4 && (test5 || test6));

			if (!fire) {
				/* distinguish between "no fire" and "unknown" */
				if (nvalid > 0) {
					FireMask[i][j] = CLASS_NO_FIRE;
					counts->NumNoFirePix++;
					}
				else {
					FireMask[i][j] = CLASS_UNKNOWN;
					counts->NumUnknownPix++;
					}

				continue;
				}


			CountAdjacentPixels(i, j, PrevScan, scan, NextScan,
				&AdjCloud, &AdjWater);


			/* override detection if sun glint is present */
			if ( (scan->pixflags[i][j].glint_level > 1) ||
				((AdjWater || nwater) && scan->pixflags[i][j].glint_level) ) {
				FireQA[i][j] |= 0x01000000;	/* bit 24 */
				counts->NumGlintRejectPix++;
				counts->NumNoFirePix++;
				FireMask[i][j] = CLASS_NO_FIRE;
				continue;
				}

			/* override detection if it's likely that this is a false
			alarm caused by excessive rejection of legitimate background
			pixels */
			if (day && !test1 && (fvalid < 0.9F) && (nbfire > 3) &&
				(BkgFireMeanT21 < 345.0F) &&
				(BkgFireMadT21 < 3.0F) && (scan->R2[i][j] > 0.15F) &&
				(T21 < (BkgFireMeanT21 + 6.0F*BkgFireMadT21))) {
				FireQA[i][j] |= 0x02000000;	/* bit 25 */
				counts->NumHotRejectPix++;
				counts->NumNoFirePix++;
				FireMask[i][j] = CLASS_NO_FIRE;
				continue;
                                }

			/* override fire pixels that are most likely to be
			false alarms caused by water-pixel contamination
			of the background */
			if (day && !test1 && nwater_cont) {
				FireQA[i][j] |= 0x04000000;	/* bit 26 */
				counts->NumCoastRejectPix++;
				counts->NumNoFirePix++;
				FireMask[i][j] = CLASS_NO_FIRE;
				continue;
				}

			C = confidence(T21, z21, DT, MeanDT, MadDT,
				AdjCloud, AdjWater, day);

			FireMask[i][j] = CLASS_FIRE_LOW + (C >= 30) + (C >= 80);
			counts->NumFirePix++;

			if (nvalid > 0)
				power = fire_power(T21, MeanT21, j);
			else
				power = 0.0F;

			st = append_fire_pixel(i + first_line, j,
				scan->lat[i][j], scan->lon[i][j],
				scan->R2[i][j],
				T21, T31, MeanT21, MeanT31, MeanDT,
				MadT21, MadT31, MadDT,
				AdjCloud, AdjWater,
				winparam, nvalid, C, power);
			if (st) return st;

			if (AdjCloud) {
				FireQA[i][j] |= 0x200000;
				counts->NumCloudAdjFirePix++;
				}

			if (AdjWater) {
				FireQA[i][j] |= 0x400000;
				counts->NumWaterAdjFirePix++;
				}
#ifndef ECS
			if (dump_fires) {
				printf("%04d,%04d %5.1f %5.1f %5.1f R:%4.2f %4.2f  "
					"B:%5.1f %5.1f %5.1f d:%4.1f %4.1f %4.1f  "
					"%1d%1d%1d%1d%1d%1d %2dw %2d\n",
					j + 1, i + first_line + 1,
					T21, T31, T21-T31, scan->R1[i][j], scan->R2[i][j],
					MeanT21, MeanT31, MeanDT,
					MadT21, MadT31, MadDT,
					test1, test2, test3, test4, test5, test6,
					L, nvalid);

				printf("    BF:%5.1f %5.1f %5.1f d:%4.1f %4.1f %4.1f %3d %4.2f\n",
					BkgFireMeanT21, BkgFireMeanT31, BkgFireMeanDT,
					BkgFireMadT21, BkgFireMadT31, BkgFireMadDT,
					nbfire, fvalid);
				}
#endif
			}
		}

	return 0;
}



static int BkgPixels(int ic, int jc, Scan *PrevScan, Scan *CurrScan, Scan *NextScan,
	float *BuffT21, float *BuffT31, float *BuffDT,
	float *BuffR1, float *BuffR2, float *BuffR7, int *nref,
	float *BkgFireBuffT21, float *BkgFireBuffT31,
	float *BkgFireBuffDT, int *nbfire,
	int *winparam, int *nwater)
/*
!C*****************************************************************************
!Description:
	Fill buffers with band 21 and 31 brightness temperatures (T21 and T31),
	and their difference (DT), for valid background pixels contained in the
	smallest square background window needed to characterize the background
	of a potential fire pixel.

	Window starts as 3 x 3 pixels in size, always excluding the two along-scan
	pixels adjacent to the candidate fire pixel, and is expanded as necessary
	until either a sufficient number of background pixels have been aquired,
	or the window becomes 21 x 21 pixels in size a sufficient number of
	background pixels still has not been obtained.  In this case, the
	characterization is deemed unsuccessful.

!Input Parameters:
	ic		scan position of potential fire pixel
	jc		sample position of potential fire pixel
	ch		arrays containing band 21 and 31 brightness temperatures
	pixflags	pixel flags array identifying valid background pixels

!Output Parameters:
	BuffT21		buffer containing T21 for valid background pixels
	BuffT31		buffer containing T31 for valid background pixels
	BuffDT		buffer containing DT for valid background pixels

	BuffR1
	BuffR2
	BuffR7
	
	nref

	BkgFireBuffT21
	BkgFireBuffT31
	BkgFireBuffDT

	nbfire

	winparam	size parameter for window needed to characterize
			background

!Input/Output Parameters:

!Returns:
	Number of valid background pixels, or 0 if an insufficient number of
	valid background pixels could be obtained.

!Revision History
	Original version 1997, L. Giglio (giglio@hades.gsfc.nasa.gov).
	May 2002 updated code to reflect version 4 algorithm changes.

!Team-unique Header:

!References and Credits

!Design Notes
	Window size is related to size parameter as:

		window size in pixels = 2 * winparam + 1

!END***************************************************************************
*/
{
	int nvalid, minvalid;
	int winsize;
	int r;
	const int rmax = 10;	/* corresponds to 21 by 21 window */

	nvalid = *nref = *nbfire = *nwater = 0;
	r = 1;		/* start at 3 x 3 ring */
	while (r <= rmax) {
		*winparam = r;
		winsize = 2*r + 1;

		GetValidBkgPixels(ic, jc, r, PrevScan, CurrScan, NextScan,
			BuffT21, BuffT31, BuffDT, &nvalid,
			BuffR1, BuffR2, BuffR7, nref,
			BkgFireBuffT21, BkgFireBuffT31, BkgFireBuffDT, nbfire,
			nwater);

		minvalid = (int) (0.25F * (winsize*winsize - 3));		/* 25% */
		if (minvalid < 8) minvalid = 8;

		if (nvalid >= minvalid) return nvalid;

		r++;
		}

	*winparam = *nref = *nbfire = *nwater = 0;
	return 0;

}



static void GetValidBkgPixels(int ic, int jc, int r,
	Scan *PrevScan, Scan *CurrScan, Scan *NextScan,
	float *BuffT21, float *BuffT31, float *BuffDT, int *nvalid,
	float *BuffR1, float *BuffR2, float *BuffR7, int *nref,
	float *BkgFireBuffT21, float *BkgFireBuffT31,
	float *BkgFireBuffDT, int *nbfire, int *nwater)
/*
!C*****************************************************************************
!Description:
	Append ch. 21, ch. 31, and ch. 21 - ch. 31 brightness temperatures of
	valid background pixels from a square "ring" of pixels to buffers.

!Input Parameters:
	ic		scan position around which ring will be centered
	jc		sample position around which ring will be centered
	r		"radius" of square ring

	PrevScan,
	CurrScan,
	NextScan	pointers to previous, current, and next scan buffers

!Output Parameters:
	BuffT21		buffer containing T21 (K) for valid background pixels
	BuffT31		buffer containing T31 (K) for valid background pixels
	BuffDT		buffer containing DT (K) for valid background pixels

!Input/Output Parameters:
	nvalid		number of valid background pixels contained in buffers
	nref
	nbfire
	nwater		number of water pixels present in background window

!Returns:

!Revision History
	Original version 1997, L. Giglio (giglio@hades.gsfc.nasa.gov).

	Modified to deal with MODIS scan bow tie overlap near scan edges,
	allowing contextual algorithm to be applied even way off nadir.
	L. Giglio, Feb. 2002.

	Modified to also tabulate values for all background fire pixels,
	Mar. 2002.

!END***************************************************************************
*/
{
	int i, j;
	int jstart, jstop;
	int jl, jr;		/* left and right samples */
	Scan *scan;

	int icorr;		/* bow-tie-corrected line index */


	if ( (jstart = jc - r) < 0) jstart = 0;
	if ( (jstop = jc + r) >= SAMPLES_1KM) jstop = SAMPLES_1KM - 1;



	/* top edge */
	i = ic - r;
	if (i < 0) {
		/* look in previous scan, handling bow tie */

		for (j = jstart; j <= jstop; j++) {
			icorr = i + 1 + iprev[j];
			if (icorr < 0) continue;

			if (PrevScan->pixflags[icorr][j].valid_bkg_pix) {
				BuffT21[*nvalid] = PrevScan->T21[icorr][j];
				BuffT31[*nvalid] = PrevScan->T31[icorr][j];
				BuffDT[*nvalid] = BuffT21[*nvalid] - BuffT31[*nvalid];
				(*nvalid)++;

				if (PrevScan->R2[icorr][j] > 0.0F) {
					BuffR1[*nref] = PrevScan->R1[icorr][j];
					BuffR2[*nref] = PrevScan->R2[icorr][j];
					BuffR7[*nref] = PrevScan->R7[icorr][j];
					(*nref)++;
					}
				}

			if (PrevScan->pixflags[icorr][j].bkg_fire) {
				BkgFireBuffT21[*nbfire] = PrevScan->T21[icorr][j];
				BkgFireBuffT31[*nbfire] = PrevScan->T31[icorr][j];
				BkgFireBuffDT[*nbfire] = BkgFireBuffT21[*nbfire] - BkgFireBuffT31[*nbfire];
				(*nbfire)++;
				}

			if (PrevScan->pixflags[icorr][j].water) (*nwater)++;
			}
		}
	else {
		/* no bowtie */
		for (j = jstart; j <= jstop; j++) {
			if (CurrScan->pixflags[i][j].valid_bkg_pix) {
				BuffT21[*nvalid] = CurrScan->T21[i][j];
				BuffT31[*nvalid] = CurrScan->T31[i][j];
				BuffDT[*nvalid] = BuffT21[*nvalid] - BuffT31[*nvalid];
				(*nvalid)++;

				if (CurrScan->R2[i][j] > 0.0F) {
					BuffR1[*nref] = CurrScan->R1[i][j];
					BuffR2[*nref] = CurrScan->R2[i][j];
					BuffR7[*nref] = CurrScan->R7[i][j];
					(*nref)++;
					}
				}

			if (CurrScan->pixflags[i][j].bkg_fire) {
				BkgFireBuffT21[*nbfire] = CurrScan->T21[i][j];
				BkgFireBuffT31[*nbfire] = CurrScan->T31[i][j];
				BkgFireBuffDT[*nbfire] = BkgFireBuffT21[*nbfire] - BkgFireBuffT31[*nbfire];
				(*nbfire)++;
				}

			if (CurrScan->pixflags[i][j].water) (*nwater)++;
			}
		}




	/* bottom edge */
	i = ic + r;
	if (i >= LINES_PER_SCAN_1KM) {
		/* look in next scan, handling bow tie */

		for (j = jstart; j <= jstop; j++) {
			icorr = i - LINES_PER_SCAN_1KM + inext[j];
			if (icorr >= LINES_PER_SCAN_1KM) continue;

			if (NextScan->pixflags[icorr][j].valid_bkg_pix) {
				BuffT21[*nvalid] = NextScan->T21[icorr][j];
				BuffT31[*nvalid] = NextScan->T31[icorr][j];
				BuffDT[*nvalid] = BuffT21[*nvalid] - BuffT31[*nvalid];
				(*nvalid)++;

				if (NextScan->R2[icorr][j] > 0.0F) {
					BuffR1[*nref] = NextScan->R1[icorr][j];
					BuffR2[*nref] = NextScan->R2[icorr][j];
					BuffR7[*nref] = NextScan->R7[icorr][j];
					(*nref)++;
					}
				}

			if (NextScan->pixflags[icorr][j].bkg_fire) {
				BkgFireBuffT21[*nbfire] = NextScan->T21[icorr][j];
				BkgFireBuffT31[*nbfire] = NextScan->T31[icorr][j];
				BkgFireBuffDT[*nbfire] = BkgFireBuffT21[*nbfire] - BkgFireBuffT31[*nbfire];
				(*nbfire)++;
				}

			if (NextScan->pixflags[icorr][j].water) (*nwater)++;
			}
		}
	else {
		/* no bowtie */
		for (j = jstart; j <= jstop; j++) {
			if (CurrScan->pixflags[i][j].valid_bkg_pix) {
				BuffT21[*nvalid] = CurrScan->T21[i][j];
				BuffT31[*nvalid] = CurrScan->T31[i][j];
				BuffDT[*nvalid] = BuffT21[*nvalid] - BuffT31[*nvalid];
				(*nvalid)++;

				if (CurrScan->R2[i][j] > 0.0F) {
					BuffR1[*nref] = CurrScan->R1[i][j];
					BuffR2[*nref] = CurrScan->R2[i][j];
					BuffR7[*nref] = CurrScan->R7[i][j];
					(*nref)++;
					}
				}

			if (CurrScan->pixflags[i][j].bkg_fire) {
				BkgFireBuffT21[*nbfire] = CurrScan->T21[i][j];
				BkgFireBuffT31[*nbfire] = CurrScan->T31[i][j];
				BkgFireBuffDT[*nbfire] = BkgFireBuffT21[*nbfire] - BkgFireBuffT31[*nbfire];
				(*nbfire)++;
				}

			if (CurrScan->pixflags[i][j].water) (*nwater)++;
			}
		}



	/* if doing 3 x 3 window, i.e. r == 1, skip pixels adjacent to the
	candidate fire pixel in the along-scan direction */
	if (r == 1) return;



	jl = jc - r;	/* left edge */
	jr =  jc + r;	/* right edge */

	/*
	value of "i"			desired value of "scan"
	------------			-----------------------
	i < 0				scan = PrevPtr

	0 <= i < LINES_PER_SCAN_1KM	scan = CurrPtr

	i >= LINES_PER_SCAN_1KM		scan = NextPtr
	*/

	/* left side */
	if (jl >= 0) {
		for (i = ic - r + 1; i < ic + r; i++) {
			/* initialize as if pixel is in current scan buffer... */
			scan = CurrScan;
			icorr = i;

			/* ...but reset if necessary */
			if (i < 0) {
				scan = PrevScan;
				icorr = i + 1 + iprev[jl];
				if (icorr < 0) continue;
				}
			else if (i >= LINES_PER_SCAN_1KM) {
				scan = NextScan;
				icorr = i - LINES_PER_SCAN_1KM + inext[jl];
				if (icorr >= LINES_PER_SCAN_1KM) continue;
				}

			if (scan->pixflags[icorr][jl].valid_bkg_pix) {
				BuffT21[*nvalid] = scan->T21[icorr][jl];
				BuffT31[*nvalid] = scan->T31[icorr][jl];
				BuffDT[*nvalid] = BuffT21[*nvalid] - BuffT31[*nvalid];
				(*nvalid)++;

				if (scan->R2[icorr][jl] > 0.0F) {
					BuffR1[*nref] = scan->R1[icorr][jl];
					BuffR2[*nref] = scan->R2[icorr][jl];
					BuffR7[*nref] = scan->R7[icorr][jl];
					(*nref)++;
					}
				}

			if (scan->pixflags[icorr][jl].bkg_fire) {
				BkgFireBuffT21[*nbfire] = scan->T21[icorr][jl];
				BkgFireBuffT31[*nbfire] = scan->T31[icorr][jl];
				BkgFireBuffDT[*nbfire] = BkgFireBuffT21[*nbfire] - BkgFireBuffT31[*nbfire];
				(*nbfire)++;
				}

			if (scan->pixflags[icorr][jl].water) (*nwater)++;
			}
		}


	/* right side */
	if (jr < SAMPLES_1KM) {
		for (i = ic - r + 1; i < ic + r; i++) {
			/* initialize as if pixel is in current scan buffer... */
			scan = CurrScan;
			icorr = i;

			/* ...but reset if necessary */
			if (i < 0) {
				scan = PrevScan;
				icorr = i + 1 + iprev[jr];
				if (icorr < 0) continue;
				}
			else if (i >= LINES_PER_SCAN_1KM) {
				scan = NextScan;
				icorr = i - LINES_PER_SCAN_1KM + inext[jr];
				if (icorr >= LINES_PER_SCAN_1KM) continue;
				}

			if (scan->pixflags[icorr][jr].valid_bkg_pix) {
				BuffT21[*nvalid] = scan->T21[icorr][jr];
				BuffT31[*nvalid] = scan->T31[icorr][jr];
				BuffDT[*nvalid] = BuffT21[*nvalid] - BuffT31[*nvalid];
				(*nvalid)++;

				if (scan->R2[icorr][jr] > 0.0F) {
					BuffR1[*nref] = scan->R1[icorr][jr];
					BuffR2[*nref] = scan->R2[icorr][jr];
					BuffR7[*nref] = scan->R7[icorr][jr];
					(*nref)++;
					}
				}

			if (scan->pixflags[icorr][jr].bkg_fire) {
				BkgFireBuffT21[*nbfire] = scan->T21[icorr][jr];
				BkgFireBuffT31[*nbfire] = scan->T31[icorr][jr];
				BkgFireBuffDT[*nbfire] = BkgFireBuffT21[*nbfire] - BkgFireBuffT31[*nbfire];
				(*nbfire)++;
				}

			if (scan->pixflags[icorr][jr].water) (*nwater)++;
			}
		}
}



static void CountAdjacentPixels(int ic, int jc,
	Scan *PrevScan, Scan *CurrScan, Scan *NextScan,
	int *AdjCloud, int *AdjWater)
/*
!C*****************************************************************************
!Description:
	Count number of cloud and water pixels adjacent to detected
	fire pixel.

!Input Parameters:
	ic
	jc

	PrevScan,
	CurrScan,
	NextScan	pointers to previous, current, and next scan buffers

!Output Parameters:
	AdjCloud
	AdjWater

!Input/Output Parameters:
	None.

!Returns:
	Nothing.

!Revision History

!Team-unique Header:

!References and Credits

!Design Notes

!END***************************************************************************
*/
{
	int i1, i2, j;
	Scan *scan1, *scan2;
	int jstart, jstop;

	*AdjCloud = *AdjWater = 0;


	/* pixels to the left and right of center pixel */
	if (jc > 0) {
		if (CurrScan->pixflags[ic][jc-1].water) (*AdjWater)++;
		if (CurrScan->pixflags[ic][jc-1].cloud) (*AdjCloud)++;
		}

	if (jc < (SAMPLES_1KM - 1)) {
		if (CurrScan->pixflags[ic][jc+1].water) (*AdjWater)++;
		if (CurrScan->pixflags[ic][jc+1].cloud) (*AdjCloud)++;
		}


	/* sets of three pixels above and below center pixel */
	if ( (jstart = jc - 1) < 0) jstart = 0;
	if ( (jstop = jc + 1) >= SAMPLES_1KM) jstop = SAMPLES_1KM - 1;

	switch (ic) {
		case 0:
			i1 = LINES_PER_SCAN_1KM - 1;
			i2 = 1;
			scan1 = PrevScan;
			scan2 = CurrScan;
			break;

		case (LINES_PER_SCAN_1KM - 1):
			i1 = ic - 1;
			i2 = 0;
			scan1 = CurrScan;
			scan2 = NextScan;
			break;

		default:
			i1 = ic - 1;
			i2 = ic + 1;
			scan1 = scan2 = CurrScan;
			break;
		}


	for (j = jstart; j <= jstop; j++) {
		if (scan1->pixflags[i1][j].water) (*AdjWater)++;
		if (scan1->pixflags[i1][j].cloud) (*AdjCloud)++;

		if (scan2->pixflags[i2][j].water) (*AdjWater)++;
		if (scan2->pixflags[i2][j].cloud) (*AdjCloud)++;
		}
}





/*
Alg. will call only call this function with n >= 6 so no error checks
on n are needed.
*/
static void compute_stats(float *x, int n, float *mean, float *mad)
{
	int	i;
	double	sum;

	/* mean */
	for (i = 0, sum = 0.0; i < n; i++) sum += (double) x[i];
	*mean = (float) (sum / n);

	/* mean absolute deviation */
	for (i = 0, sum = 0.0; i < n; i++)
		sum += (double) fabsf(x[i] - *mean);

	*mad = (float) (sum / n);
}




static int SunGlint(float R1, float R2, float R7, float ts, float tv, float phi,
	float *GlintAng)
/*
!C*****************************************************************************
!Description:
	Perform sun-glint test specified in Giglio et al. 2003.

!Input Parameters:
	R1	channel 1 (0.64 micron) 1 km aggregated reflectance
	R2	channel 2 (0.86 micron) 1 km aggregated reflectance
	R7	channel 7 (2.1 micron)  1 km aggregated reflectance
	ts	solar zenith angle ("theta s") in degrees
	tv	view zenith angle ("theta v") in degrees
	phi	relative azimuth angle in degrees

!Output Parameters:
	GlintAng	"glint" angle (degrees)

!Input/Output Parameters:
	None.

!Returns:
	0 if sun glint absent, otherwise glint level from 1 (lowest)
	to 3 (highest)

!Revision History
	Original version 1997, L. Giglio (giglio@hades.gsfc.nasa.gov).
	Used mutated Li et al. tests.

	L. Giglio, 14 May 1998
	Glint angle increased to from 30 to 40 degrees as specified by Y. Kaufman.

	L. Giglio, 26 Feb. 2001
	R1 and R2 thresholds decreased from 0.30 to 0.28.

	L. Giglio, 28 Feb. 2002
	Changed glint angle calculation to match that in Giglio, et al. 2002.
	Old method used E. Prins calculation but this suffered from a sign
	error.

	August 2002
	Added band 7 and modified reflective channel thresholds.

!Team-unique Header:

!References and Credits

!Design Notes

!END***************************************************************************
*/
{
	float cos_tr;			/* cosine of reflected sun angle "theta r" */
	float tr;

	cos_tr = cosf(D2R * tv) * cosf(D2R * ts) -
		sinf(D2R * tv) * sinf(D2R * ts) * cosf(D2R * phi);

	tr = R2D * acosf(cos_tr);

	*GlintAng = tr;

	/* automatically assume glint for very small glint angles */
	if (tr < 2.0F) return 3;
        if ((tr < 8.0F) && (R1 > 0.10F) && (R2 > 0.20F) && (R7 > 0.12F)) return 2;
	if (tr < 12.0F) return 1;

	return 0;
}




static float fixang(float ang)
/*
!C*****************************************************************************
!Description:
	Map angle in range [-540,+540] to range [-180,+180].

!Input Parameters:
	ang	angle (degrees)

!Output Parameters:
	None.

!Input/Output Parameters:
	None.

!Returns:
	Returns angle mapped to range [-180,+180]

!Revision History

!Team-unique Header:

!References and Credits

!Design Notes

!END***************************************************************************
*/
{
	if (ang > +180.0F) return ang - 360.0F;
	if (ang <= -180.0F) return 360.0F + ang;
	return ang;
}




static int confidence(float T21, float z21,
	float DT, float MeanDT, float MadDT,
	int AdjCloud, int AdjWater, int day)
/*
!C*****************************************************************************
!Description:
	Experimental function to compute confidence of detected
	fire pixels.

!Input Parameters:
	T21
	z21
	DT
	MeanDT
	MadDT
	AdjCloud	number of cloud pixels adjacent to fire pixel
	AdjWater	number of water pixels adjacent to fire pixel
	day		day/night flag (non-zero --> day)

!Output Parameters:
	None.

!Input/Output Parameters:
	None.

!Returns:
	Returns detection confidence in percent.

!Revision History
	Original version April 2001, L. Giglio.

	L. Giglio, Jul. 2004
	Properly handle case of failed background characterization.

	L. Giglio, April 2006
	Slightly altered daytime thresholds for C1, C4, and C5.

!END***************************************************************************
*/
{
	float C1, C2, C3, C4, C5;	/* individual confidence tests */
	float C;
	float zDT;

	const float eps = 1.0e-10F;	/* tiny value used to prevent division by zero */



	/* special case: no background statistics available */
	if (z21 < eps) {
		/*
		In this case only the conservative absolute threshold
		test was passed, so T21 will be at least 360 K
		(= DayParams.thresh_T21) during the day and 320 K
		(= NightParams.thresh_T21) at night.
		Because these thresholds are so conservative, I set
		the ramp function parameters so that pixels
		meeting even these minimum thresholds will have a
		confidence of 50%, and rapidly increase for
		brightness temperatures above these thresholds.
		*/
		if (day)
			C = S(T21, 350.0F, 370.0F);
		else
			C = S(T21, 310.0F, 330.0F);

		/* convert to percent */
		return (int) (100.0F * C);
		}


	zDT = (DT - MeanDT) / (MadDT + eps);

	/* keep these consistent with algorithm parameters */
	C2 = S(z21, 3.0F, 6.0F);
	C3 = S(zDT, 3.5F, 6.0F);

	if (day) {
		C1 = S(T21, 310.0F, 360.0F);
		C4 = 1.0F - S((float) AdjCloud, 0.0F, 4.0F);
		C5 = 1.0F - S((float) AdjWater, 0.0F, 4.0F);
		C = powf(C1 * C2 * C3 * C4 * C5, 0.2F);
		}
	else {
		C1 = S(T21, 305.0F, 320.0F);
		C = powf(C1 * C2 * C3, 0.333333F);
		}

	/* convert to percent */
	return (int) (100.0F * C);
}



/*
Evaluate unit ramp function having specified thresholds.
Returns:
	0		(x < a)
	(x-a)/(b-a)	(a <= x <= b)
	1		(x > b)
*/
static float S(float x, float a, float b)
{
	if (x <= a) return 0.0F;
	if (x >= b) return 1.0F;

	return (x - a) / (b - a);
}




static float fire_power(float T21, float MeanT21, int sample)
/*
!C*****************************************************************************
!Description:
	Implementation of Yoram Kaufman's active fire power calculation.
	See ATBD for details.

!Input Parameters:

!Output Parameters:
	Returns fire power in MW.

!Revision History:
	Original version August 1996, L. Giglio (giglio@hades.gsfc.nasa.gov).
	Revised April 2006.

!Team-unique Header:

!Design Notes
	Intermediate calculation is done in double precision since the
	difference between two very large numbers is calculated.  The final
	result is then cast into single precision.
!END***************************************************************************
*/
{
	double power;
	const double C = 4.34e-19;
	double FRP;


	if (T21 > MeanT21) {
		power = C * (pow((double) T21, 8.0) -
			pow((double) MeanT21, 8.0));
		FRP = power * pixel_area(sample);
		}
	else
		power = FRP = 0.0;

	return (float) FRP;
}



/* does NOT handle missing data in input arrays; make sure there isn't any when calling */
static void compute_ndvi(float *R1, float *R2, int n, float *NDVI)
{
	int i;

	for (i = 0; i < n; i++)
		NDVI[i] = (R2[i] - R1[i]) / (R2[i] + R1[i]);
}


/* returns number of background pixels that are very likely to contain water */
static int bkg_water_pix(float *R2, float *R7, float *NDVI, int n)
{
	int i;
	int nwater;

	for (i = nwater = 0; i < n; i++)
		if ((R2[i] < 0.15F) && (R7[i] < 0.05F) && (NDVI[i] < 0.0F)) nwater++;

	return nwater;
}



/*
Calculate approximate area of 1-km MODIS pixel at specified position within
scan.
*/
static float pixel_area(int sample)
{
	const double h = 705.0;		/* Terra/Aqua orbit altitude [km] */
	const double p = 1.0;		/* nadir pixel resolution [km] */
	double r, s;
	double theta, cos_theta;
	double temp, sqrt_temp;
	double DS, DT, area;


	r = EARTH_RADIUS + h;		/* [km] */
	s = p / h;			/* [rad] */

	theta = sample*s + 0.5*s - (0.5*SAMPLES_1KM)*s;
	cos_theta = cos(theta);

	temp = pow(EARTH_RADIUS/r,2.0) - pow(sin(theta),2.0);
	sqrt_temp = sqrt(temp);

	DS = EARTH_RADIUS * s * (cos_theta/sqrt_temp - 1.0);
	DT = r*s*(cos_theta - sqrt_temp);
	area = DS * DT;

	return area;
}

