/*
Utility to read multiple MODIS Level 2 fire product HDF files (MOD14 and MYD14)
and dump selected information about each fire pixel detected in the granule to
standard output.  Information includes latitude, longitude, acquisition date,
channel 21/22 brightness temperature, fire radiative power (FRP), and detection
confidence.

L. Giglio, August 2000

Revised March and April 2001 to change format and include date field for
UMD rapid-response system.

Revised May 2002 to include pixel dimension fields in output.

Revised February 2003 to include detection confidence in output.

Revised by J. Kendall in June 2007 to include fire radiative power (FRP)
field in output. Note: The value for FRP is extracted directly from SDS
FP_power. For the Collection 5 MOD14/MYD14 product this is valid; for
Collection 4 it is not -- the dumped value must be multiplied by the pixel area
to obtain the actual FRP.  See the "MODIS Collection 4 Active Fire Product
User's Guide" for details.

Usage:
	fireloc <MOD14/MYD14 HDF file>

Compilation:
	gcc -ansi -o fireloc fireloc.c -I$HDFINC -L$HDFLIB -lmfhdf \
	-ldf -lz -ljpeg

Set HDFINC and HDFLIB environment variables to the repective paths of the HDF
header files and object libraries.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mfhdf.h"



int read_fire_pixel_table(int32 sd_id);
int ReadSDS(int32 sd_id, char *name, int32 esize, int32 etype, void *data);
int DumpFireInfo(char *infile);
int get_num_fires(int32 sd_id, int32 *nfire);
int ReadHDF(int32 sd_id, char **corebuff, char **archbuff);
void PixelSize(int16 sample, double *Sh, double *Sl);



main(int argc, char **argv)
{
	int i;
	char *buff_core, *buff_arch;
	int nmeta;


	if (argc < 2) {
		(void) fputs("Usage: fireloc <MODIS L2 fire HDF files>\n",
			stderr);
		exit(1);
		}

	for (i = 1; i < argc; i++) {
		if (DumpFireInfo(argv[i]))
			(void) fprintf(stderr, "Error reading file \"%s\".\n", argv[i]);
		}

	exit(0);
}



int DumpFireInfo(char *infile)
{
	int32 sd_id;
	int32 st1, st2, st3, st4, st5, st6;

	int32 i;
	int32 nfires;
	int32 type;		/* expected data type */

	float32	*lat, *lon;	/* degrees */
	float32 *T21;		/* K */
	float32 *FRP;		/* fire radiative power */
	int16 *sample;
	uint8 *conf;
	double Sh, Sl;

	size_t nbytes;

	if ((sd_id = SDstart(infile, DFACC_READ)) == FAIL) return 1;

	/* unable to read global HDF attribute, so get # fire pixels this way */
	if (get_num_fires(sd_id, &nfires)) return 2;

	/* bail out if no fire pixels */
	if (!nfires) {
		(void) SDend(sd_id);
		return 0;
		}

	nbytes = nfires * sizeof(float32);
	lat = (float32 *) malloc(nbytes);
	lon = (float32 *) malloc(nbytes);
	T21 = (float32 *) malloc(nbytes);
	FRP = (float32 *) malloc(nbytes);
	sample = (int16 *) malloc(nfires * sizeof(int16));
	conf = (uint8 *) malloc(nfires * sizeof(uint8));

	if (!lat || !lon || !T21 || !sample || !conf) return 3;

	type = DFNT_FLOAT32;
	st1 = ReadSDS(sd_id, "latitude", nfires, type, lat);
	if (st1) st1 = ReadSDS(sd_id, "FP_latitude", nfires, type, lat);
	st2 = ReadSDS(sd_id, "longitude", nfires, type, lon);
	if (st2) st2 = ReadSDS(sd_id, "FP_longitude", nfires, type, lon);
	st3 = ReadSDS(sd_id, "T21", nfires, type, T21);
	if (st3) st3 = ReadSDS(sd_id, "FP_T21", nfires, type, T21);
	st4 = ReadSDS(sd_id, "FP_sample", nfires, DFNT_INT16, sample);
	st5 = ReadSDS(sd_id, "FP_confidence", nfires, DFNT_UINT8, conf);
	st6 = ReadSDS(sd_id, "FP_power", nfires, type, FRP);

	if (st1 || st2 || st3 || st4 || st5 || st6) {
		free(conf);
		free(sample);
		free(T21);
		free(FRP);
		free(lon);
		free(lat);
		(void) SDend(sd_id);
		return 4;
		}

	for (i = 0; i < nfires; i++) {
		PixelSize(sample[i], &Sh, &Sl);

		printf("%.3f,%.3f,%.1f,%.1f,%.1f,%d,%.1f\n", lat[i], lon[i], T21[i],
			Sh, Sl, conf[i], FRP[i]);
		}

	free(conf);
	free(sample);
	free(T21);
	free(lon);
	free(lat);
	free(FRP);

	(void) SDend(sd_id);

	return 0;
}




int ReadSDS(int32 sd_id, char *name, int32 esize, int32 etype, void *data)
/*
! 
C*********************************************************************** 
******
!Description:
	Read 1-D SDS having specified name from HDF file.

!Input Parameters:
	sd_id
	name		SDS name
	esize		expected array length
	etype		expected array data type

!Output Parameters:
	data		array to store data

!Returns:
	Zero if no errors, non-zero otherwise.

!Revision History:
	Original version Sept. 1999, L. Giglio (giglio@hades.gsfc.nasa.gov).

!Team-unique Header:
         References and Credits
         Design Notes
! 
END********************************************************************* 
******
*/
{
	int32 sds_id;
	int32 sds_index;

	int32 rank, data_type, nattr;
	char sds_name[MAX_NC_NAME];
	int32 dim_sizes[MAX_VAR_DIMS];

	int32 start[1];
	int32 edges[1];

/* puts(name); */

	if ((sds_index = SDnametoindex(sd_id, name)) == FAIL) return 1;
	if ((sds_id = SDselect(sd_id, sds_index)) == FAIL) return 2;
	if (SDgetinfo(sds_id, sds_name, &rank, dim_sizes, &data_type, &nattr)  
== FAIL) return 3;

	/* check rank */
	if (rank != 1) return 4;

	/* check length */
	if (dim_sizes[0] != esize) return 5;

	/* check type */
	if (data_type != etype) return 6;

	start[0] = start[1] = 0;
	edges[0] = esize;
	if (SDreaddata(sds_id, start, NULL, edges, data) == FAIL) return 7;

	if (SDendaccess(sds_id) == FAIL) return 8;

	return 0;
}



int get_num_fires(int32 sd_id, int32 *nfire)
/*
! 
C*********************************************************************** 
******
!Description:
	Determine number of fire pixels in MODIS Level 2 fire product file.

!Input Parameters:
	sd_id

!Output Parameters:
	nfire		number of fire pixels

!Returns:
	Zero if no errors, non-zero otherwise.

!Revision History:
	Original version Sept. 1999, L. Giglio (giglio@hades.gsfc.nasa.gov).

!Team-unique Header:
         References and Credits
         Design Notes
! 
END********************************************************************* 
******
*/
{
	int32 sds_id;
	int32 sds_index;

	int32 rank, data_type, nattr;
	char sds_name[MAX_NC_NAME];
	int32 dim_sizes[MAX_VAR_DIMS];

	if ((sds_index = SDnametoindex(sd_id, "FP_line")) == FAIL) {
		/* look for old name */
		if ((sds_index = SDnametoindex(sd_id, "line")) == FAIL) return 1;
		}
	if ((sds_id = SDselect(sd_id, sds_index)) == FAIL) return 2;
	if (SDgetinfo(sds_id, sds_name, &rank, dim_sizes, &data_type, &nattr)  
== FAIL) return 3;

	*nfire = dim_sizes[0];

	if (SDendaccess(sds_id) == FAIL) return 4;

	return 0;
}







/*
Ingest core and archive metadata strings from EOS-style HDF file.
*/
int ReadHDF(int32 sd_id, char **corebuff, char **archbuff)
{
	int32 icore, iarch;
	char dummy[32];
	int32 attr_cnt, attr_type;
	long bytes_core, bytes_arch;
	char *buff_core, *buff_arch;

	*corebuff = *archbuff = (char *) NULL;

	if ((icore = SDfindattr(sd_id, "CoreMetadata.0")) == FAIL) {
		SDend(sd_id);
		return 2;
		}

	if ((iarch = SDfindattr(sd_id, "ArchiveMetadata.0")) == FAIL) {
		SDend(sd_id);
		return 3;
		}

	if (SDattrinfo(sd_id, icore, dummy, &attr_type, &attr_cnt) == FAIL) {
		SDend(sd_id);
		return 4;
		}

	/* add extra byte for terminating null */
	bytes_core = (long) DFKNTsize(attr_type) * (long) attr_cnt + 1L;


	if (SDattrinfo(sd_id, iarch, dummy, &attr_type, &attr_cnt) == FAIL) {
		SDend(sd_id);
		return 5;
		}

	/* add extra byte for terminating null */
	bytes_arch = (long) DFKNTsize(attr_type) * (long) attr_cnt + 1L;


	buff_core = (char *) malloc((size_t) bytes_core);
	buff_arch = (char *) malloc((size_t) bytes_arch);

	if (!buff_core || !buff_arch) {
		if (buff_core) free((void *) buff_core);
		if (buff_arch) free((void *) buff_arch);
		SDend(sd_id);
		return 6;
		}

	if (SDreadattr(sd_id, icore, (void *) buff_core) == FAIL) {
		free((void *) buff_core);
		free((void *) buff_arch);
		SDend(sd_id);
		return 7;
		}


	if (SDreadattr(sd_id, iarch, (void *) buff_arch) == FAIL) {
		free((void *) buff_core);
		free((void *) buff_arch);
		SDend(sd_id);
		return 8;
		}

	/* add terminating nulls */
	buff_core[bytes_core - 1] = buff_arch[bytes_arch - 1] = '\0';

	*corebuff = buff_core;
	*archbuff = buff_arch;

	return 0;
}



/*
Quick hack to compute MODIS pixel size as function of sample number.
Quick-and-dirty empirical fit to laboriously computed values.

sample should be between 0 and 1353, inclusive.
*/
void PixelSize(int16 sample, double *Sh, double *Sl)
{
int i;
static double as[7] = {0.979416, -8.31278e-09, 4.67105e-06,
7.14429e-14, -1.52269e-11, -1.24302e-19, 5.07407e-17};
static double at[] = {0.998637, -1.38590e-09, 9.03223e-07,
4.14437e-15, 2.81020e-12};

i = sample - 677;

/* along-scan */
*Sh = as[0] + i * (as[1] + i * (as[2] + i * (as[3] + i * (as[4] + i * (as[5] + i * as[6])))));

/* along-track */
*Sl = at[0] + i * (at[1] + i * (at[2] + i * (at[3] + i * at[4])));
}

