/*
!C******************************************************************

!Description: 
	Functions for dealing with the "fire pixel table" written into
	the MODIS Level 2 fire product.  The fire pixel table is
	actually a collection of many one-dimensional SDS's containing
	information about each fire pixel identified within a granule.

	Originally the fire pixel table was supposed to be a single HDF
	vdata containing multiple fields, an HDF construct intended for
	storing tabular data.  However, because of a horrible limitation
	imposed upon vdatas by the HDF library (specifically, the sequence
	for opening and closing HDF files is mutually exclusive for SDS's
	and vdatas), and limitations within vdatas themselves, multiple SDS's
	had to be used instead.


The following two headings are included so ECS software doesn't
complain.  They are in fact meaningless at the module level.

!Input Parameters:
!Output Parameters:

!Revision History:
	Original version, August 1999, L. Giglio (giglio@hades.gsfc.nasa.gov).

        Revised 30-NOV-99: writes all data to a buffer instead of directly to
	the SDS (J. Ray).

	Completely rewritten Feb. 2001, L. Giglio.

!Team-unique Header:
	See any of the other fifty dozen functions in this code for this
	information.

!References and Credits:
	See any of the other fifty dozen functions in this code for this
	information.

!Design Notes
	Some kludginess is present in this module to work around the way
	in which the Land HDF function library ("lhdf") creates SDS's.
	Specifically, "lhdf" doesn't create SDS's until the first
	line of data is written with Lhdf_Put().  Because I (Giglio) wanted
	the fire pixel table SDS's to follow the fire mask and algorithm QA
	SDS's in the fire product (since most users are only interested
	in these two SDS's), the functions in this module were written
	so that creation of the fire pixel table was delayed until _after_
	Lhdf_Put() was called for the first line in each granule.

	The proper sequence of public functions within this module is:

		1. Call new_fire_pixel_table().

		2. Call append_fire_pixel() for each fire pixel
		identified within a granule.

		3. Call write_fire_pixel_table() after each granule has
		been processed.

	This should be done for each L2 fire product granule generated.

		4. Call free_fire_pixel_table() after all granules have been
		processed.

!END*************************************************************************/


#include "mfhdf.h"
#include "firetab.h"
#include "error.h"


extern int verbose;



/* structure to contain info. about each SDS comprising fire pixel table */
typedef struct {
	char	*name;
	int32	type;		/* HDF data type */
	char	*desc;		/* description (long name) */
	char	*unit;
	void	*data;
	} SDS;



enum {FPTAB_LINE, FPTAB_SAMPLE, FPTAB_LAT, FPTAB_LON, FPTAB_R2,
	FPTAB_T21, FPTAB_T31,
	FPTAB_MEAN21, FPTAB_MEAN31, FPTAB_MEDDT,
	FPTAB_SDEV21, FPTAB_SDEV31, FPTAB_SDEVDT,
	FPTAB_POWER, FPTAB_ADJCLOUD, FPTAB_ADJWATER,
	FPTAB_WINSIZE, FPTAB_NVALID, FPTAB_CONF,
	NUM_SDS};

/*
These are the separate SDS's comprising the "fire pixel table".
Order of table entries must match that of above enumeration.
*/
static SDS sds[NUM_SDS] = {
	{"FP_line",	DFNT_INT16,	"granule line of fire pixel",	(char *) NULL,	(void *) NULL},
	{"FP_sample",	DFNT_INT16,	"granule sample of fire pixel",	(char *) NULL,	(void *) NULL},
	{"FP_latitude",	DFNT_FLOAT32,	"latitude of fire pixel",	"degrees",	(void *) NULL},
	{"FP_longitude",	DFNT_FLOAT32,	"longitude of fire pixel",	"degrees",	(void *) NULL},

	{"FP_R2",		DFNT_FLOAT32,	"channel 2 reflectance "
					"of fire pixel",		(char *) NULL,	(void *) NULL},

	{"FP_T21",		DFNT_FLOAT32,	"channel 21/22 brightness "
					"temperature of fire pixel",	"kelvins",	(void *) NULL},

	{"FP_T31",		DFNT_FLOAT32,	"channel 31 brightness "
					"temperature of fire pixel",	"kelvins",	(void *) NULL},


 	{"FP_MeanT21",	DFNT_FLOAT32,	"channel 21/22 brightness "
					"temperature of background",	"kelvins", (void *) NULL},

	{"FP_MeanT31",	DFNT_FLOAT32,	"channel 31 brightness "
					"temperature of background",	"kelvins", (void *) NULL},

	{"FP_MeanDT",	DFNT_FLOAT32,	"mean background brightness "
					"temperature difference",	"kelvins", (void *) NULL},

	{"FP_MAD_T21",	DFNT_FLOAT32,	"background channel 21/22 brightness "
					"temperature mean absolute deviation", "kelvins", (void *) NULL},

	{"FP_MAD_T31",	DFNT_FLOAT32,	"background channel 31 brightness "
					"temperature mean absolute deviation", "kelvins", (void *) NULL},

	{"FP_MAD_DT",	DFNT_FLOAT32,	"background brightness temperature "
					"difference mean absolute deviation", "kelvins", (void *) NULL},

	{"FP_power",	DFNT_FLOAT32,	"fire radiative power",	"MW",	(void *) NULL},
	{"FP_AdjCloud",	DFNT_UINT8,	"number of adjacent cloud pixels", (char *) NULL, (void *) NULL},
	{"FP_AdjWater",	DFNT_UINT8,	"number of adjacent water pixels", (char *) NULL, (void *) NULL},
	{"FP_WinSize",	DFNT_UINT8,	"background window size",	(char *) NULL,	(void *) NULL},
	{"FP_NumValid",	DFNT_INT16,	"number of valid background pixels",	(char *) NULL,	(void *) NULL},
	{"FP_confidence", DFNT_UINT8,	"detection confidence",		"%",	(void *) NULL}
	};





static long nfire = 0L;
static long nalloc = 0L;


#define ALLOC_ELEMS 1000


int append_fire_pixel(int line, int sample, float lat, float lon,
	float R2, float T21, float T31, float MeanT21, float MeanT31,
	float MedianDT, float MAD_T21, float MAD_T31, float MAD_DT,
	int AdjCloud, int AdjWater, int winparam, int nvalid, int confidence,
	float power)
{
	int i;
	uint8	*bptr;
	int16	*iptr;
	float32	*fptr;
	size_t bytes;

	if (nfire >= nalloc) {
		nalloc += ALLOC_ELEMS;

#ifdef DEBUG
	if (verbose) printf("FPTAB: reallocating %d elements.\n", nalloc);
#endif

		for (i = 0; i < NUM_SDS; i++) {
			bytes = DFKNTsize(sds[i].type) * nalloc;

			if (!(sds[i].data = realloc(sds[i].data, bytes))) return error(ERR_F_MEM);
			}
		}


	iptr = (int16 *) sds[FPTAB_LINE].data;
	iptr[nfire] = (int16) line;

	iptr = (int16 *) sds[FPTAB_SAMPLE].data;
	iptr[nfire] = (int16) sample;

	fptr = (float32 *) sds[FPTAB_LAT].data;
	fptr[nfire] = lat;

	fptr = (float32 *) sds[FPTAB_LON].data;
	fptr[nfire] = lon;

	fptr = (float32 *) sds[FPTAB_R2].data;
	fptr[nfire] = R2;

	fptr = (float32 *) sds[FPTAB_T21].data;
	fptr[nfire] = T21;

	fptr = (float32 *) sds[FPTAB_T31].data;
	fptr[nfire] = T31;

	fptr = (float32 *) sds[FPTAB_MEAN21].data;
	fptr[nfire] = MeanT21;

	fptr = (float32 *) sds[FPTAB_MEAN31].data;
	fptr[nfire] = MeanT31;

	fptr = (float32 *) sds[FPTAB_MEDDT].data;
	fptr[nfire] = MedianDT;

	fptr = (float32 *) sds[FPTAB_SDEV21].data;
	fptr[nfire] = MAD_T21;

	fptr = (float32 *) sds[FPTAB_SDEV31].data;
	fptr[nfire] = MAD_T31;

	fptr = (float32 *) sds[FPTAB_SDEVDT].data;
	fptr[nfire] = MAD_DT;

	fptr = (float32 *) sds[FPTAB_POWER].data;
	fptr[nfire] = power;

	bptr = (uint8 *) sds[FPTAB_ADJCLOUD].data;
	bptr[nfire] = (uint8) AdjCloud;

	bptr = (uint8 *) sds[FPTAB_ADJWATER].data;
	bptr[nfire] = (uint8) AdjWater;

	iptr = (int16 *) sds[FPTAB_NVALID].data;
	iptr[nfire] = (int16) nvalid;

	bptr = (uint8 *) sds[FPTAB_WINSIZE].data;
	if (winparam)
		bptr[nfire] = (uint8) (2 * winparam + 1);
	else
		bptr[nfire] = (uint8) 0;

	bptr = (uint8 *) sds[FPTAB_CONF].data;
	bptr[nfire] = (uint8) confidence;

	nfire++;

	return 0;
}



void free_fire_pixel_table(void)
{
	int i;

	for (i = 0; i < NUM_SDS; i++) if (sds[i].data) free(sds[i].data);

	nfire = nalloc = 0L;
}



void new_fire_pixel_table(void)
{
	nfire = 0L;
}




int write_fire_pixel_table(int32 sd_id)
/*
!C*****************************************************************************

!Description:
	Create and writes the SDS's comprising the fire pixel table
	which has been stored internally up to this point.

!Input Parameters:
	sd_id  HDF SD ID.

	Uses private variables "nfire" and array "sds".

!Output Parameters:
	None.

!Input/Ouput Parameters:
	None.

!Returns:
	Returns 0 if successful, non-zero otherwise.

!Revision History:
	Original version, L. Giglio (August 1999)
	Modified Feb. 2001 (L. Giglio) for V3 fire code.

!Team-unique Header:
	See any of the other fifty dozen functions in this code for this
	information.

!References and Credits:
	See any of the other fifty dozen functions in this code for this
	information.

!Design Notes


!END
****************************************************************************/
{
	int i;
	int32 sds_id;
	int32 dim_sizes[1];
	int32 start[1], edges[1];
	int32 dim_id;


	dim_sizes[0] = edges[0] = nfire;
	start[0] = 0;

	for (i = 0; i < NUM_SDS; i++) {
		if ((sds_id = SDcreate(sd_id, sds[i].name, sds[i].type, 1, dim_sizes)) == FAIL)
			return error(ERR_HDF);
		if ((dim_id = SDgetdimid(sds_id, 0)) == FAIL)
			return error(ERR_HDF);
		if (SDsetdimname(dim_id, "number_of_active_fires") == FAIL)
			return error(ERR_HDF);
		if (SDsetdatastrs(sds_id, sds[i].desc, sds[i].unit, NULL, NULL) == FAIL)
			return error(ERR_HDF);

		/* only write SDS if we have at least one fire pixel */
		if (nfire)
			if (SDwritedata(sds_id, start, NULL, edges, sds[i].data) == FAIL)
				return error(ERR_HDF);

		if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);
		}
 
	return 0;
}

