/*
!C******************************************************************
!Description:
	Functions to handle input from and output to HDF files.

!Revision History:
!Team-unique Header:
!References and Credits:

	L. Giglio, Jan. 2001

!Design Notes:

	Although it would be better if it did not, this software was required
	by ECS to make use of the Hierarchical Data Format software created
	and distributed by the National Center for Supercomputing
	Applications.

!END*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mfhdf.h"

#include "const.h"
#include "types.h"
#include "tb.h"
#include "fire.h"
#include "detfire.h"
#include "firetab.h"
#include "io.h"
#include "process.h"
#include "error.h"
#include "misc.h"
#include "version.h"



#ifndef ECS
#include <unistd.h>
extern int write_tb;

static int init_thermal_common_output(int32 sds_id, int channel, char *dimname1,
	char *dimname2);
static int identify_satellite(int32 sd_id_mod03, enum Satellite *sat);
static char *DimStr(int32 rank, int32 *dim_sizes);
#endif


static void unscale(uint16 Lscaled[LINES_PER_SCAN_1KM][SAMPLES_1KM],
	float L[LINES_PER_SCAN_1KM][SAMPLES_1KM], float scale, float offset,
	float Lsat, UBYTE dead[LINES_PER_SCAN_1KM]);
static void unscale_angle(int16 iang[LINES_PER_SCAN_1KM][SAMPLES_1KM],
	float ang[LINES_PER_SCAN_1KM][SAMPLES_1KM], float scale);

static void make_coarse(uint8 coarse_fire[LINES_PER_SCAN_5KM][SAMPLES_5KM]);
static char *satstr(enum Satellite sat);


/* output buffers for the fire product */
uint8	FireMask[LINES_PER_SCAN_1KM][SAMPLES_1KM];
uint32	FireQA[LINES_PER_SCAN_1KM][SAMPLES_1KM];



#define EMISS_1KM_BANDS 16
#define REF_250M_BANDS 2
#define REF_500M_BANDS 5


/* 1 km emissive bands radiance scales and offsets */
static float radscale[EMISS_1KM_BANDS], radoff[EMISS_1KM_BANDS];

/* aggregated 250 m reflectance scales and offsets */
static float refscale[REF_250M_BANDS], refoff[REF_250M_BANDS];

/* aggregated 500 m reflectance scales and offsets */
static float refscale500[REF_500M_BANDS], refoff500[REF_500M_BANDS];

extern int verbose;
extern int compress_sds;



int InitOutputHDF(int32 sd_id_fire, int32 sd_id_coarse, int nscans, HDF_OutputInfo *hinfo)
{
	int32 sds_id;

	int32 dim_sizes[2];
	int32 dim_id;
	char *dimname1, *dimname2;

	int32 n_values;
#ifndef ECS
	int32 ulen;
#endif
	uint8 ubuff[2];

	char *units;
	char *legend;

	HDF_CHUNK_DEF cdef;


	/* ---------- Level 2 Fire Product ---------- */
	hinfo->sd_id_fire = sd_id_fire;

	dim_sizes[0] = nscans * 10;
	dim_sizes[1] = SAMPLES_1KM;

	dimname1 = "number_of_scan_lines";
	dimname2 = "pixels_per_scan_line";


	hinfo->sds_index_mask = 0;
	if ((sds_id = SDcreate(sd_id_fire, "fire mask", DFNT_UINT8, 2, dim_sizes)) == FAIL)
		return error(ERR_HDF);

	if (compress_sds) {
		cdef.chunk_lengths[0] = cdef.comp.chunk_lengths[0] = LINES_PER_SCAN_1KM;
		cdef.chunk_lengths[1] = cdef.comp.chunk_lengths[1] = SAMPLES_1KM;
		cdef.comp.comp_type = COMP_CODE_DEFLATE;
		cdef.comp.cinfo.deflate.level = 4;
		if (SDsetchunk(sds_id, cdef, HDF_CHUNK | HDF_COMP) == FAIL)
			return error(ERR_HDF);
		}

	if ((dim_id = SDgetdimid(sds_id, 0)) == FAIL)
		return error(ERR_HDF);
	if (SDsetdimname(dim_id, dimname1) == FAIL) return error(ERR_HDF);
	if ((dim_id = SDgetdimid(sds_id, 1)) == FAIL) return error(ERR_HDF);
	if (SDsetdimname(dim_id, dimname2) == FAIL) return error(ERR_HDF);

	ubuff[0] = 0;
	ubuff[1] = 9;
	n_values = 2;
	if (SDsetattr(sds_id, "valid_range", DFNT_UINT8, n_values, ubuff))
		return error(ERR_HDF);

#ifdef JUNK
	units = "bit field";
	n_values = strlen(units);
	if (SDsetattr(sds_id, "units", DFNT_CHAR8, n_values, units)) return (ERR_HDF);
#endif


legend = "Classes:\n"
	"0 missing input data\n"
	"1 not processed (obsolete)\n"
	"2 not processed (obsolete)\n"
	"3 water\n4 cloud\n5 non-fire\n"
	"6 unknown\n7 fire (low confidence)\n"
	"8 fire (nominal confidence)\n9 fire (high confidence)";

	n_values = strlen(legend);
	if (SDsetattr(sds_id, "legend", DFNT_CHAR8, n_values, legend))
		return error(ERR_HDF);

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


	hinfo->sds_index_qa = 1;
	if ((sds_id = SDcreate(sd_id_fire, "algorithm QA", DFNT_UINT32, 2, dim_sizes)) == FAIL)
		return error(ERR_HDF);

	if (compress_sds) {
		cdef.chunk_lengths[0] = cdef.comp.chunk_lengths[0] = LINES_PER_SCAN_1KM;
		cdef.chunk_lengths[1] = cdef.comp.chunk_lengths[1] = SAMPLES_1KM;
		cdef.comp.comp_type = COMP_CODE_DEFLATE;
		cdef.comp.cinfo.deflate.level = 4;
		if (SDsetchunk(sds_id, cdef, HDF_CHUNK | HDF_COMP) == FAIL)
			return error(ERR_HDF);
		}

	if ((dim_id = SDgetdimid(sds_id, 0)) == FAIL) return error(ERR_HDF);
	if (SDsetdimname(dim_id, dimname1) == FAIL) return error(ERR_HDF);
	if ((dim_id = SDgetdimid(sds_id, 1)) == FAIL) return error(ERR_HDF);
	if (SDsetdimname(dim_id, dimname2) == FAIL) return error(ERR_HDF);

	units = "bit field";
	n_values = strlen(units);
	if (SDsetattr(sds_id, "units", DFNT_CHAR8, n_values, units))
		return error(ERR_HDF);

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


#ifndef ECS
	if (write_tb) {
		units = "kelvins";
		ulen = strlen(units);

		hinfo->sds_index_T21 = 2;
		if ((sds_id = SDcreate(sd_id_fire, "T21", DFNT_UINT16, 2, dim_sizes)) == FAIL) return (ERR_HDF);
		if (init_thermal_common_output(sds_id, 21, dimname1, dimname2)) return (ERR_HDF);
		if (SDendaccess(sds_id) == FAIL) return (ERR_HDF);

		hinfo->sds_index_T31 = 3;
		if ((sds_id = SDcreate(sd_id_fire, "T31", DFNT_UINT16, 2, dim_sizes)) == FAIL) return (ERR_HDF);
		if (init_thermal_common_output(sds_id, 31, dimname1, dimname2)) return (ERR_HDF);
		if (SDendaccess(sds_id) == FAIL) return (ERR_HDF);

		hinfo->sds_index_T32 = 4;
		if ((sds_id = SDcreate(sd_id_fire, "T32", DFNT_UINT16, 2, dim_sizes)) == FAIL) return (ERR_HDF);
		if (init_thermal_common_output(sds_id, 32, dimname1, dimname2)) return (ERR_HDF);
		if (SDendaccess(sds_id) == FAIL) return (ERR_HDF);
		}
	else
		hinfo->sds_index_T21 = hinfo->sds_index_T31 =
			hinfo->sds_index_T32 = -1;
#endif

	/* ---------- Level 2 Coarse-Resolution Fire Product ---------- */
	if ( (hinfo->sd_id_coarse = sd_id_coarse) >= 0) {
		dim_sizes[0] = nscans * 2;	/* two 5-km lines per scan */
		dim_sizes[1] = SAMPLES_5KM;

		dimname1 = "number_of_coarse_resolution_scan_lines";
		dimname2 = "coarse_resolution_pixels_per_scan_line";

		hinfo->sds_index_coarse_mask = 0;
		if ((sds_id = SDcreate(sd_id_coarse, "CoarseFireCounts", DFNT_UINT8, 2, dim_sizes)) == FAIL)
			return error(ERR_HDF);

		if ((dim_id = SDgetdimid(sds_id, 0)) == FAIL)
			return error(ERR_HDF);
		if (SDsetdimname(dim_id, dimname1) == FAIL) return error(ERR_HDF);
		if ((dim_id = SDgetdimid(sds_id, 1)) == FAIL) return error(ERR_HDF);
		if (SDsetdimname(dim_id, dimname2) == FAIL) return error(ERR_HDF);

		if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);

		units = "degrees";
		n_values = strlen(units);


		hinfo->sds_index_coarse_lat = 1;
		if ((sds_id = SDcreate(sd_id_coarse, "Latitude", DFNT_FLOAT32, 2, dim_sizes)) == FAIL)
			return error(ERR_HDF);

		if ((dim_id = SDgetdimid(sds_id, 0)) == FAIL)
			return error(ERR_HDF);
		if (SDsetdimname(dim_id, dimname1) == FAIL) return error(ERR_HDF);
		if ((dim_id = SDgetdimid(sds_id, 1)) == FAIL) return error(ERR_HDF);
		if (SDsetdimname(dim_id, dimname2) == FAIL) return error(ERR_HDF);
		if (SDsetattr(sds_id, "units", DFNT_CHAR8, n_values, units))
			return error(ERR_HDF);

		if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


		hinfo->sds_index_coarse_lon = 2;
		if ((sds_id = SDcreate(sd_id_coarse, "Longitude", DFNT_FLOAT32, 2, dim_sizes)) == FAIL)
			return error(ERR_HDF);

		if ((dim_id = SDgetdimid(sds_id, 0)) == FAIL)
			return error(ERR_HDF);
		if (SDsetdimname(dim_id, dimname1) == FAIL) return error(ERR_HDF);
		if ((dim_id = SDgetdimid(sds_id, 1)) == FAIL) return error(ERR_HDF);
		if (SDsetdimname(dim_id, dimname2) == FAIL) return error(ERR_HDF);
		if (SDsetattr(sds_id, "units", DFNT_CHAR8, n_values, units))
			return error(ERR_HDF);

		if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);
		}
	else
		hinfo->sds_index_coarse_mask = hinfo->sds_index_coarse_lat =
			hinfo->sds_index_coarse_lon = -1;

	return 0;
}



#ifndef ECS

static int init_thermal_common_output(int32 sds_id, int channel, char *dimname1, char *dimname2)
{
	int32 dim_id;
	int32 ulen, llen;
	char *units;
	char long_name[64];
	int16 missing_value = 0;
	float32 scale_factor = 0.1F;

	units = "kelvins";
	(void) sprintf(long_name, "channel %d brightness temperature", channel);

	ulen = strlen(units);
	llen = strlen(long_name);

	if ((dim_id = SDgetdimid(sds_id, 0)) == FAIL) return (ERR_HDF);
	if (SDsetdimname(dim_id, dimname1) == FAIL) return (ERR_HDF);
	if ((dim_id = SDgetdimid(sds_id, 1)) == FAIL) return (ERR_HDF);
	if (SDsetdimname(dim_id, dimname2) == FAIL) return (ERR_HDF);
	if (SDsetattr(sds_id, "long_name", DFNT_CHAR8, llen, long_name)) return (ERR_HDF);
	if (SDsetattr(sds_id, "units", DFNT_CHAR8, ulen, units)) return (ERR_HDF);
	if (SDsetattr(sds_id, "missing_value", DFNT_INT16, 1, &missing_value)) return (ERR_HDF);
	if (SDsetattr(sds_id, "scale_factor", DFNT_FLOAT32, 1, &scale_factor)) return (ERR_HDF);

	return 0;
}

#endif



/*
	sd_id_mod02	MOD021KM 1B product
	sd_id_mod03	MOD03 geolocation product
*/
int InitInputHDF(int32 sd_id_mod02, int32 sd_id_mod03, HDF_InputInfo *hinfo)
{
	int32 sds_id;
	int32 rank, data_type, nattr;
	char sds_name[MAX_NC_NAME];
	int32 dim_sizes[MAX_VAR_DIMS];
	int32 attr_index;
	int32 n_values;
	char attr_name[MAX_NC_NAME];
	int st;

	int i;
	int8 dead_list[490];


	hinfo->sd_id_mod02 = sd_id_mod02;
	hinfo->sd_id_mod03 = sd_id_mod03;


	/* Could parse band names to set these.  Hard code for now. */
	hinfo->index21 = 1;
	hinfo->index22 = 2;
	hinfo->index31 = 10;
	hinfo->index32 = 11;
	hinfo->index1 = 0;
	hinfo->index2 = 1;
	hinfo->index7 = 4;


	/* ---------- MOD021KM ---------- */

	/* 1 km emissive bands */
	if ((hinfo->sds_index_emiss = SDnametoindex(sd_id_mod02, "EV_1KM_Emissive")) == FAIL)
		return error(ERR_HDF);
	if ((sds_id = SDselect(sd_id_mod02, hinfo->sds_index_emiss)) == FAIL)
		return error(ERR_HDF);
	if (SDgetinfo(sds_id, sds_name, &rank, dim_sizes, &data_type, &nattr) == FAIL)
		return error(ERR_HDF);

#ifndef ECS
	if (verbose) printf("name = %24s, rank = %2d, data type = %2d  dims: %s\n",
		sds_name, rank, data_type, DimStr(rank, dim_sizes));
#endif

	/* check rank and data type */
	if (rank != 3) return error(ERR_DATA_RANK);
	if (data_type != DFNT_UINT16) return error(ERR_DATA_TYPE);


	/* read radiance scales */
	if ((attr_index = SDfindattr(sds_id, "radiance_scales")) == FAIL)
		return error(ERR_HDF);
	if (SDattrinfo(sds_id, attr_index, attr_name, &data_type, &n_values))
		return error(ERR_HDF);
	if (data_type != DFNT_FLOAT32) return error(ERR_DATA_TYPE);
	if (SDreadattr(sds_id, attr_index, (void *) radscale))
		return error(ERR_HDF);


	/* read radiance offsets */
	if ((attr_index = SDfindattr(sds_id, "radiance_offsets")) == FAIL)
		return error(ERR_HDF);
	if (SDattrinfo(sds_id, attr_index, attr_name, &data_type, &n_values))
		return error(ERR_HDF);
	if (data_type != DFNT_FLOAT32) return error(ERR_DATA_TYPE);
	if (SDreadattr(sds_id, attr_index, (void *) radoff))
		return error(ERR_HDF);

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);

	/*
	Number of image lines should be a multiple of 10 to match the ten detectors
	acquiring data during each scan.
	*/
	if (dim_sizes[1] % 10) return error(ERR_IMAGE_LINES);

	/* ten lines per full MODIS scan */
	hinfo->nscans = dim_sizes[1] / 10;


	/* aggregated 250 m bands */
	if ((hinfo->sds_index_ref = SDnametoindex(sd_id_mod02, "EV_250_Aggr1km_RefSB")) == FAIL)
		return error(ERR_HDF);
	if ((sds_id = SDselect(sd_id_mod02, hinfo->sds_index_ref)) == FAIL)
		return error(ERR_HDF);
	if (SDgetinfo(sds_id, sds_name, &rank, dim_sizes, &data_type, &nattr) == FAIL)
		return error(ERR_HDF);

#ifndef ECS
	if (verbose) printf("name = %24s, rank = %2d, data type = %2d  dims: %s\n",
		sds_name, rank, data_type, DimStr(rank, dim_sizes));
#endif

	/* check rank and data type */
	if (rank != 3) return error(ERR_DATA_RANK);
	if (data_type != DFNT_UINT16) return error(ERR_DATA_TYPE);

	/* read 250 m reflectance scales */
	if ((attr_index = SDfindattr(sds_id, "reflectance_scales")) == FAIL)
		return error(ERR_HDF);
	if (SDattrinfo(sds_id, attr_index, attr_name, &data_type, &n_values))
		return error(ERR_HDF);
	if (data_type != DFNT_FLOAT32) return error(ERR_DATA_TYPE);
	if (SDreadattr(sds_id, attr_index, (void *) refscale))
		return error(ERR_HDF);


	/* read 250 m reflectance offsets */
	if ((attr_index = SDfindattr(sds_id, "reflectance_offsets")) == FAIL)
		return error(ERR_HDF);
	if (SDattrinfo(sds_id, attr_index, attr_name, &data_type, &n_values))
		return error(ERR_HDF);
	if (data_type != DFNT_FLOAT32) return error(ERR_DATA_TYPE);
	if (SDreadattr(sds_id, attr_index, (void *) refoff)) return error(ERR_HDF);

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);



	/* aggregated 500 m bands */
	if ((hinfo->sds_index_ref500 = SDnametoindex(sd_id_mod02, "EV_500_Aggr1km_RefSB")) == FAIL)
		return error(ERR_HDF);
	if ((sds_id = SDselect(sd_id_mod02, hinfo->sds_index_ref500)) == FAIL)
		return error(ERR_HDF);
	if (SDgetinfo(sds_id, sds_name, &rank, dim_sizes, &data_type, &nattr) == FAIL)
		return error(ERR_HDF);

#ifndef ECS
	if (verbose) printf("name = %24s, rank = %2d, data type = %2d  dims: %s\n",
		sds_name, rank, data_type, DimStr(rank, dim_sizes));
#endif

	/* check rank and data type */
	if (rank != 3) return error(ERR_DATA_RANK);
	if (data_type != DFNT_UINT16) return error(ERR_DATA_TYPE);

	/* read 500 m reflectance scales */
	if ((attr_index = SDfindattr(sds_id, "reflectance_scales")) == FAIL)
		return error(ERR_HDF);
	if (SDattrinfo(sds_id, attr_index, attr_name, &data_type, &n_values))
		return error(ERR_HDF);
	if (data_type != DFNT_FLOAT32) return error(ERR_DATA_TYPE);
	if (SDreadattr(sds_id, attr_index, (void *) refscale500))
		return error(ERR_HDF);


	/* read 500 m reflectance offsets */
	if ((attr_index = SDfindattr(sds_id, "reflectance_offsets")) == FAIL)
		return error(ERR_HDF);
	if (SDattrinfo(sds_id, attr_index, attr_name, &data_type, &n_values))
		return error(ERR_HDF);
	if (data_type != DFNT_FLOAT32) return error(ERR_DATA_TYPE);
	if (SDreadattr(sds_id, attr_index, (void *) refoff500)) return error(ERR_HDF);

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);




	/* dead detector list (global attribute) */
	if ((attr_index = SDfindattr(sd_id_mod02, "Dead Detector List")) == FAIL)
		return error(ERR_HDF);
	if (SDattrinfo(sd_id_mod02, attr_index, attr_name, &data_type, &n_values))
		return error(ERR_HDF);

#ifndef ECS
	if (verbose) printf("name = %24s, data type = %2d  n_values: %d\n",
		attr_name, data_type, n_values);
#endif
	if (n_values != 490) return error(ERR_DATA_SIZE);
	if (data_type != DFNT_INT8) return error(ERR_DATA_TYPE);
	if (SDreadattr(sd_id_mod02, attr_index, (void *) dead_list)) return error(ERR_HDF);

	/*
	Order of bands within dead detector list is like this:

	1, 2, 3, .. , 13lo, 13hi, 14lo, 14hi, 15, .. , 36.

	first 2 * 40 = 80 values:	bands 1 and 2;	elements 0-79
	next 5 * 20 = 100 values:	bands 3-7;	elements 80-179
	next 15 * 10 = 150 values:	bands 8-20;	elements 180-329
	next 10 values:			band 21;	elements 330-339
	next 10 values:			band 22;	elements 340-349
	next 8 * 10 = 80 values:	bands 23-30;	elements 350-429
	next 10 values: 		band 31;	elements 430-439
	next 5 * 10 = 50 values:	bands 32-36;	elements 440-489
	*/

	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		hinfo->dead21[i] = dead_list[i + 330];
		hinfo->dead22[i] = dead_list[i + 340];
		hinfo->dead31[i] = dead_list[i + 430];
		hinfo->dead32[i] = dead_list[i + 440];
		}

#ifndef ECS
	if (verbose) {
		puts("band 21, 22, 31, 32 dead detector list");

		for (i = 0; i < LINES_PER_SCAN_1KM; i++)
			printf("%d %1d %1d %1d %1d\n", i, hinfo->dead21[i],
				hinfo->dead22[i], hinfo->dead31[i], hinfo->dead32[i]);

		puts("");
		}
#endif


	/* ---------- MOD03 ---------- */

	/* geolocation */
	if ((hinfo->sds_index_lat = SDnametoindex(sd_id_mod03, "Latitude")) == FAIL)
		return error(ERR_HDF);
	if ((sds_id = SDselect(sd_id_mod03, hinfo->sds_index_lat)) == FAIL)
		return error(ERR_HDF);
	if (SDgetinfo(sds_id, sds_name, &rank, dim_sizes, &data_type, &nattr) == FAIL)
		return error(ERR_HDF);

#ifndef ECS
	if (verbose) printf("name = %24s, rank = %2d, data type = %2d  dims: %s\n",
		sds_name, rank, data_type, DimStr(rank, dim_sizes));
#endif

	if (dim_sizes[0] != 10*hinfo->nscans) return error(ERR_NLINES_MISMATCH);


	if (data_type != DFNT_FLOAT32) return error(ERR_DATA_TYPE);

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


	if ((hinfo->sds_index_lon = SDnametoindex(sd_id_mod03, "Longitude")) == FAIL)
		return error(ERR_HDF);
	if ((sds_id = SDselect(sd_id_mod03, hinfo->sds_index_lon)) == FAIL)
		return error(ERR_HDF);
	if (SDgetinfo(sds_id, sds_name, &rank, dim_sizes, &data_type, &nattr) == FAIL)
		return error(ERR_HDF);
#ifndef ECS
	if (verbose) printf("name = %24s, rank = %2d, data type = %2d  dims: %s\n",
		sds_name, rank, data_type, DimStr(rank, dim_sizes));
#endif

	if (data_type != DFNT_FLOAT32) return error(ERR_DATA_TYPE);

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


	if ((hinfo->sds_index_viewzen = SDnametoindex(sd_id_mod03, "SensorZenith")) == FAIL)
		return error(ERR_HDF);
	if ((sds_id = SDselect(sd_id_mod03, hinfo->sds_index_viewzen)) == FAIL)
		return error(ERR_HDF);
	if (SDgetinfo(sds_id, sds_name, &rank, dim_sizes, &data_type, &nattr) == FAIL)
		return error(ERR_HDF);
#ifndef ECS
	if (verbose) printf("name = %24s, rank = %2d, data type = %2d  dims: %s\n",
		sds_name, rank, data_type, DimStr(rank, dim_sizes));
#endif

	if (data_type != DFNT_INT16) return error(ERR_DATA_TYPE);

	if ( (st = ReadScaleFactor(sds_id, &hinfo->scale_viewzen)) ) return st;

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


/* FIX: can pass null ptr instead of sds_name */
	if ((hinfo->sds_index_solzen = SDnametoindex(sd_id_mod03, "SolarZenith")) == FAIL)
		return error(ERR_HDF);
	if ((sds_id = SDselect(sd_id_mod03, hinfo->sds_index_solzen)) == FAIL)
		return error(ERR_HDF);
	if (SDgetinfo(sds_id, sds_name, &rank, dim_sizes, &data_type, &nattr) == FAIL)
		return error(ERR_HDF);
#ifndef ECS
	if (verbose) printf("name = %24s, rank = %2d, data type = %2d  dims: %s\n",
		sds_name, rank, data_type, DimStr(rank, dim_sizes));
#endif
	if (data_type != DFNT_INT16) return error(ERR_DATA_TYPE);

	if ( (st = ReadScaleFactor(sds_id, &hinfo->scale_solzen)) ) return st;

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


	if ((hinfo->sds_index_viewaz = SDnametoindex(sd_id_mod03, "SensorAzimuth")) == FAIL)
		return error(ERR_HDF);
	if ((sds_id = SDselect(sd_id_mod03, hinfo->sds_index_viewaz)) == FAIL)
		return error(ERR_HDF);
	if (SDgetinfo(sds_id, sds_name, &rank, dim_sizes, &data_type, &nattr) == FAIL)
		return error(ERR_HDF);
#ifndef ECS
	if (verbose) printf("name = %24s, rank = %2d, data type = %2d  dims: %s\n",
		sds_name, rank, data_type, DimStr(rank, dim_sizes));
#endif
	if (data_type != DFNT_INT16) return error(ERR_DATA_TYPE);

	if ( (st = ReadScaleFactor(sds_id, &hinfo->scale_viewaz)) ) return st;

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);



	if ((hinfo->sds_index_solaz = SDnametoindex(sd_id_mod03, "SolarAzimuth")) == FAIL)
		return error(ERR_HDF);
	if ((sds_id = SDselect(sd_id_mod03, hinfo->sds_index_solaz)) == FAIL)
		return error(ERR_HDF);
	if (SDgetinfo(sds_id, sds_name, &rank, dim_sizes, &data_type, &nattr) == FAIL)
		return error(ERR_HDF);
#ifndef ECS
	if (verbose) printf("name = %24s, rank = %2d, data type = %2d  dims: %s\n",
		sds_name, rank, data_type, DimStr(rank, dim_sizes));
#endif
	if (data_type != DFNT_INT16) return error(ERR_DATA_TYPE);

	if ( (st = ReadScaleFactor(sds_id, &hinfo->scale_solaz)) ) return st;

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);



	if ((hinfo->sds_index_landsea = SDnametoindex(sd_id_mod03, "Land/SeaMask")) == FAIL)
		return error(ERR_HDF);
	if ((sds_id = SDselect(sd_id_mod03, hinfo->sds_index_landsea)) == FAIL)
		return error(ERR_HDF);
	if (SDgetinfo(sds_id, sds_name, &rank, dim_sizes, &data_type, &nattr) == FAIL)
		return error(ERR_HDF);
#ifndef ECS
	if (verbose) printf("name = %24s, rank = %2d, data type = %2d  dims: %s\n",
		sds_name, rank, data_type, DimStr(rank, dim_sizes));
#endif
	if (data_type != DFNT_UINT8) return error(ERR_DATA_TYPE);

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


	/*
	Use MOD03/MYD03 geolocation to identify satellite for stand-alone
	(non-ECS) code.
	*/
#ifndef ECS
	if ( (st = identify_satellite(sd_id_mod03, &hinfo->sat)) ) return st;
#else
	hinfo->sat = SAT_UNKNOWN;
#endif


#ifndef ECS
	if (verbose) {
		puts("radiance scales and offsets:");
		for (i = 0; i < 36; i++) printf("%2d  %e  %e\n", i, radscale[i], radoff[i]);
		}
#endif


	return 0;
}



int ReadScaleFactor(int32 sds_id, float32 *scale_factor)
{
	int32 attr_index;
	int32 data_type, n_values;
	char attr_name[MAX_NC_NAME];
	float64	dtemp;

	if ((attr_index = SDfindattr(sds_id, "scale_factor")) == FAIL)
		return error(ERR_HDF);

	if (SDattrinfo(sds_id, attr_index, attr_name, &data_type, &n_values))
		return error(ERR_HDF);

	if (data_type != DFNT_FLOAT64) return error(ERR_DATA_TYPE);

	if (SDreadattr(sds_id, attr_index, (void *) &dtemp))
		return error(ERR_HDF);

	/*
	Geolocation angles are needlessly stored with 64 bits of precision.
	Cast them into floats here so that subsequent unscaling can be done
	with faster single precision.
	*/
	*scale_factor = (float) dtemp;

	return 0;
}



#ifndef ECS
static char *DimStr(int32 rank, int32 *dim_sizes)
{
	int i;
	char numstr[8];
	static char s[128];

	s[0] = '\0';
	for (i = 0; i < rank; i++) {
		(void) sprintf(numstr, "%6d", dim_sizes[i]);
		(void) strcat(s, numstr);
		}

	return s;
}
#endif



/*
Scan pointer is passed in case one of the buffered quantities needs to be
added to L2 fire product.  It is currently only used in in unwarped (i.e. non-ECS)
mode to write brightness temperature arrays.
*/
int WriteScan(HDF_OutputInfo *hinfo, int iscan, Scan *scan)
{
	int32 sds_id;
	int32 start[2];
	int32 edges[2];

	int i, j;
	int icoarse, jcoarse;

#ifndef ECS
	uint16 ubuff[LINES_PER_SCAN_1KM][SAMPLES_1KM];
#endif

	uint8 coarse_fire[LINES_PER_SCAN_5KM][SAMPLES_5KM];
	float32 coarse_lat[LINES_PER_SCAN_5KM][SAMPLES_5KM];
	float32 coarse_lon[LINES_PER_SCAN_5KM][SAMPLES_5KM];


	start[0] = LINES_PER_SCAN_1KM * iscan;
	edges[0] = LINES_PER_SCAN_1KM;

	start[1] = 0;
	edges[1] = SAMPLES_1KM;

	if ((sds_id = SDselect(hinfo->sd_id_fire, hinfo->sds_index_mask)) == FAIL)
		return error(ERR_HDF);
	if (SDwritedata(sds_id, start, NULL, edges, (void *) FireMask) == FAIL)
		return error(ERR_HDF);
	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);

	if ((sds_id = SDselect(hinfo->sd_id_fire, hinfo->sds_index_qa)) == FAIL)
		return error(ERR_HDF);
	if (SDwritedata(sds_id, start, NULL, edges, (void *) FireQA) == FAIL)
		return error(ERR_HDF);
	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


#ifndef ECS
	if (write_tb) {
		/* band 21/22 hybrid */
		for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
			for (j = 0; j < SAMPLES_1KM; j++) {
				if (scan->T21[i][j] > 0.0F)
					ubuff[i][j] = (uint16) (10.0F * scan->T21[i][j]);
				else
					ubuff[i][j] = 0;
				}
			}

		if ((sds_id = SDselect(hinfo->sd_id_fire, hinfo->sds_index_T21)) == FAIL)
			return error(ERR_HDF);
		if (SDwritedata(sds_id, start, NULL, edges, (void *) ubuff) == FAIL)
			return error(ERR_HDF);
		if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


		/* band 31 */
		for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
			for (j = 0; j < SAMPLES_1KM; j++) {
				if (scan->T31[i][j] > 0.0F)
					ubuff[i][j] = (uint16) (10.0F * scan->T31[i][j]);
				else
					ubuff[i][j] = 0;
				}
			}

		if ((sds_id = SDselect(hinfo->sd_id_fire, hinfo->sds_index_T31)) == FAIL)
			return error(ERR_HDF);
		if (SDwritedata(sds_id, start, NULL, edges, (void *) ubuff) == FAIL)
			return error(ERR_HDF);
		if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


		/* band 32 */
		for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
			for (j = 0; j < SAMPLES_1KM; j++) {
				if (scan->T31[i][j] > 0.0F)
					ubuff[i][j] = (uint16) (10.0F * scan->T32[i][j]);
				else
					ubuff[i][j] = 0;
				}
			}

		if ((sds_id = SDselect(hinfo->sd_id_fire, hinfo->sds_index_T32)) == FAIL)
			return error(ERR_HDF);
		if (SDwritedata(sds_id, start, NULL, edges, (void *) ubuff) == FAIL)
			return error(ERR_HDF);
		if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);
		}
#endif

	if (hinfo->sd_id_coarse >= 0) {
		make_coarse(coarse_fire);

		start[0] = LINES_PER_SCAN_5KM * iscan;
		edges[0] = LINES_PER_SCAN_5KM;

		start[1] = 0;
		edges[1] = SAMPLES_5KM;

		if ((sds_id = SDselect(hinfo->sd_id_coarse, hinfo->sds_index_coarse_mask)) == FAIL)
			return error(ERR_HDF);
		if (SDwritedata(sds_id, start, NULL, edges, (void *) coarse_fire) == FAIL)
			return error(ERR_HDF);
		if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


		/* sample latitude and longitude buffers at center of 5km pixels */
		for (icoarse = 0; icoarse < LINES_PER_SCAN_5KM; icoarse++) {
			i = icoarse * 5 + 2;	/* center of 5 x 5 box of 1 km pixels */

			for (jcoarse = 0; jcoarse < SAMPLES_5KM; jcoarse++) {
				j = jcoarse * 5 + 2;	/* center of 5 x 5 box of 1 km pixels */

				coarse_lat[icoarse][jcoarse] = scan->lat[i][j];
				coarse_lon[icoarse][jcoarse] = scan->lon[i][j];
				}
			}


		if ((sds_id = SDselect(hinfo->sd_id_coarse, hinfo->sds_index_coarse_lat)) == FAIL)
			return error(ERR_HDF);
		if (SDwritedata(sds_id, start, NULL, edges, (void *) coarse_lat) == FAIL)
			return error(ERR_HDF);
		if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


		if ((sds_id = SDselect(hinfo->sd_id_coarse, hinfo->sds_index_coarse_lon)) == FAIL)
			return error(ERR_HDF);
		if (SDwritedata(sds_id, start, NULL, edges, (void *) coarse_lon) == FAIL)
			return error(ERR_HDF);
		if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);
		}

	return 0;
}



/*
Generate a buffer's worth of the coarse-resolution fire product.
*/
static void make_coarse(uint8 coarse_fire[LINES_PER_SCAN_5KM][SAMPLES_5KM])
{
	int i, j;
	int icoarse, jcoarse;

	UBYTE ncloud[LINES_PER_SCAN_5KM][SAMPLES_5KM];
	UBYTE nwater[LINES_PER_SCAN_5KM][SAMPLES_5KM];
	UBYTE nvalid[LINES_PER_SCAN_5KM][SAMPLES_5KM];

	for (icoarse = 0; icoarse < LINES_PER_SCAN_5KM; icoarse++) {
		for (jcoarse = 0; jcoarse < SAMPLES_5KM; jcoarse++) {
			coarse_fire[icoarse][jcoarse] = 0;
			ncloud[icoarse][jcoarse] = nwater[icoarse][jcoarse] = 0;
			nvalid[icoarse][jcoarse] = 0;
			}
		}


	/* count fire, cloud, water, and valid data pixels in scan */
	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		icoarse = i / 5;		/* map 1 km pixels into 5 km */

		for (j = 0; j < SAMPLES_1KM; j++) {
			jcoarse = j / 5;	/* map 1 km pixels into 5 km */

			if (isfire(FireMask[i][j])) {
				++coarse_fire[icoarse][jcoarse];
				continue;
				}

			if (FireMask[i][j] == CLASS_CLOUD) ++ncloud[icoarse][jcoarse];
			if (FireMask[i][j] == CLASS_WATER) ++nwater[icoarse][jcoarse];

			if (FireMask[i][j] >= CLASS_WATER) ++nvalid[icoarse][jcoarse];
			}
		}


	/*
	Flag presence of cloud and water pixels in coarse-resolution pixels.

	If no valid data was in a coarse-resolution pixel we'll handle that
	here as well by filling the coarse product with the missing data code 255.
	*/
	for (icoarse = 0; icoarse < LINES_PER_SCAN_5KM; icoarse++) {
		for (jcoarse = 0; jcoarse < SAMPLES_5KM; jcoarse++) {
			if (!nvalid[icoarse][jcoarse]) {
				coarse_fire[icoarse][jcoarse] = 255;
				continue;
				}

			/*
			bits 0-4 used for fire pixels counts
			bit 5 used for cloud flag
			bit 6 used for water flag
			*/
			if (ncloud[icoarse][jcoarse]) coarse_fire[icoarse][jcoarse] |= 32;
			if (nwater[icoarse][jcoarse]) coarse_fire[icoarse][jcoarse] |= 64;
			}
		}
}



/*
Read a single 1 km MODIS scan from required input files.

	hinfo		HDF input file info.
	iscan		scan number (0 <= iscan < hinfo.nscans)
	scan		Scan structure in which data should be stored

Returns 0 if no errors, non-zero error code otherwise.
*/
int ReadScan(HDF_InputInfo *hinfo, int iscan, Scan *scan)
{
	int32 sds_id;
	int32 start[3];
	int32 edges[3];

/* FIX: someday make these arrays a union */
	uint16 buff[LINES_PER_SCAN_1KM][SAMPLES_1KM];
	float fbuff[LINES_PER_SCAN_1KM][SAMPLES_1KM];
	int16 ibuff[LINES_PER_SCAN_1KM][SAMPLES_1KM];

	float Lsat21, Lsat31;
	const float Lsat22 = 2.076F;	/* 330.8 K */
	const float Lsat32 = 25.10F;	/* 400 K */

	/* linear correction coefficients for spectral response */
	static float corr21[2] = {1.00009F, -0.05167F};
	static float corr22[2] = {1.0001F, -0.05332F};
	static float corr31[2] = {1.00046F, -0.09968F};
	static float corr32[2] = {1.00063F, -0.13162F};

	/* dummy dead detector array for aggregated bands (no dead detectors) */
	static UBYTE dummy_dead[LINES_PER_SCAN_1KM] = {0,0,0,0,0,0,0,0,0,0};


	if (hinfo->sat == SAT_AQUA) {
		/* Aqua MODIS */
		Lsat21 = 90.5F;		/* 504 K */
		Lsat31 = 18.62F;	/* 353 K */
		}
	else {
		/* Terra MODIS */
		Lsat21 = 85.44F;	/* 500 K */
		Lsat31 = 29.17F;	/* 400 K */
		}

	edges[0] = 1;		/* read one band at a time */

	start[1] = 10 * iscan;
	edges[1] = LINES_PER_SCAN_1KM;	/* read 10 scan lines (i.e. one complete scan) */

	start[2] = 0;
	edges[2] = SAMPLES_1KM;


	/* ---------- MOD02 emissive bands ---------- */

	if ((sds_id = SDselect(hinfo->sd_id_mod02, hinfo->sds_index_emiss)) == FAIL)
		return error(ERR_HDF);

	/* channel 21 radiance */
	start[0] = hinfo->index21;
	if (SDreaddata(sds_id, start, NULL, edges, (void *) buff) == FAIL)
		return error(ERR_HDF);
	unscale(buff, scan->L21, radscale[hinfo->index21], radoff[hinfo->index21],
		Lsat21, hinfo->dead21);
	compute_tb(scan->L21, scan->T21, 3.96F);
	correct_tb(scan->T21, corr21);

	/* channel 22 radiance */
	start[0] = hinfo->index22;
	if (SDreaddata(sds_id, start, NULL, edges, (void *) buff) == FAIL)
		return error(ERR_HDF);
	unscale(buff, scan->L22, radscale[hinfo->index22], radoff[hinfo->index22],
		Lsat22, hinfo->dead22);
	compute_tb(scan->L22, scan->T22, 3.96F);
	correct_tb(scan->T22, corr22);

	/* channel 31 radiance */
	start[0] = hinfo->index31;
	if (SDreaddata(sds_id, start, NULL, edges, (void *) buff) == FAIL)
		return error(ERR_HDF);
	unscale(buff, fbuff, radscale[hinfo->index31], radoff[hinfo->index31],
		Lsat31, hinfo->dead31);
	compute_tb(fbuff, scan->T31, 11.01F);
	correct_tb(scan->T31, corr31);


	/* channel 32 radiance (used for internal cloud mask only) */
	start[0] = hinfo->index32;
	if (SDreaddata(sds_id, start, NULL, edges, (void *) buff) == FAIL)
		return error(ERR_HDF);
	unscale(buff, fbuff, radscale[hinfo->index32], radoff[hinfo->index32],
		Lsat32, hinfo->dead32);
	compute_tb(fbuff, scan->T32, 12.03F);
	correct_tb(scan->T32, corr32);

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);



	/* ---------- MOD02 aggregated 250 m reflective bands ---------- */

	/* NB. Here we're reading reflectance cos(theta). */
	if ((sds_id = SDselect(hinfo->sd_id_mod02, hinfo->sds_index_ref)) == FAIL)
		return error(ERR_HDF);

	/* channel 1 non-normalized reflectance */
	start[0] = hinfo->index1;
	if (SDreaddata(sds_id, start, NULL, edges, (void *) buff) == FAIL)
		return error(ERR_HDF);
	unscale(buff, scan->R1, refscale[hinfo->index1], refoff[hinfo->index1],
		100.0F, dummy_dead);


	/* channel 2 non-normalized reflectance */
	start[0] = hinfo->index2;
	if (SDreaddata(sds_id, start, NULL, edges, (void *) buff) == FAIL)
		return error(ERR_HDF);
	unscale(buff, scan->R2, refscale[hinfo->index2], refoff[hinfo->index2],
		100.0F, dummy_dead);

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);



	/* ---------- MOD02 aggregated 500 m reflective bands ---------- */

	/* NB. Here we're reading reflectance cos(theta). */
	if ((sds_id = SDselect(hinfo->sd_id_mod02, hinfo->sds_index_ref500)) == FAIL)
		return error(ERR_HDF);

	/* channel 7 non-normalized reflectance */
	start[0] = hinfo->index7;
	if (SDreaddata(sds_id, start, NULL, edges, (void *) buff) == FAIL)
		return error(ERR_HDF);
	unscale(buff, scan->R7, refscale500[hinfo->index7], refoff500[hinfo->index7],
		100.0F, dummy_dead);

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);



	/* ---------- MOD03 geolocation and angles ---------- */

	start[0] = 10 * iscan;
	edges[0] = LINES_PER_SCAN_1KM;	/* read 10 scan lines (i.e. one complete scan) */

	start[1] = 0;
	edges[1] = SAMPLES_1KM;

	if ((sds_id = SDselect(hinfo->sd_id_mod03, hinfo->sds_index_lat)) == FAIL)
		return error(ERR_HDF);
	if (SDreaddata(sds_id, start, NULL, edges, (void *) scan->lat) == FAIL)
		return error(ERR_HDF);
	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);

	if ((sds_id = SDselect(hinfo->sd_id_mod03, hinfo->sds_index_lon)) == FAIL)
		return error(ERR_HDF);
	if (SDreaddata(sds_id, start, NULL, edges, (void *) scan->lon) == FAIL)
		return error(ERR_HDF);
	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);

	if ((sds_id = SDselect(hinfo->sd_id_mod03, hinfo->sds_index_viewzen)) == FAIL)
		return error(ERR_HDF);
	if (SDreaddata(sds_id, start, NULL, edges, (void *) ibuff) == FAIL)
		return error(ERR_HDF);
	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);
	unscale_angle(ibuff, scan->ViewZenAng, hinfo->scale_viewzen);

	if ((sds_id = SDselect(hinfo->sd_id_mod03, hinfo->sds_index_solzen)) == FAIL)
		return error(ERR_HDF);
	if (SDreaddata(sds_id, start, NULL, edges, (void *) ibuff) == FAIL)
		return error(ERR_HDF);
	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);
	unscale_angle(ibuff, scan->SolZenAng, hinfo->scale_solzen);

	if ((sds_id = SDselect(hinfo->sd_id_mod03, hinfo->sds_index_viewaz)) == FAIL)
		return error(ERR_HDF);
	if (SDreaddata(sds_id, start, NULL, edges, (void *) ibuff) == FAIL)
		return error(ERR_HDF);
	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);
	unscale_angle(ibuff, scan->ViewAzAng, hinfo->scale_viewaz);

	if ((sds_id = SDselect(hinfo->sd_id_mod03, hinfo->sds_index_solaz)) == FAIL)
		return error(ERR_HDF);
	if (SDreaddata(sds_id, start, NULL, edges, (void *) ibuff) == FAIL)
		return error(ERR_HDF);
	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);
	unscale_angle(ibuff, scan->SolAzAng, hinfo->scale_solaz);

	if ((sds_id = SDselect(hinfo->sd_id_mod03, hinfo->sds_index_landsea)) == FAIL)
		return error(ERR_HDF);
	if (SDreaddata(sds_id, start, NULL, edges, (void *) scan->landsea) == FAIL)
		return error(ERR_HDF);
	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


	/* ---------- divide reflectance cos(theta) by cos(theta) ---------- */

	compute_reflectances(scan->R1, scan->R2, scan->R7, scan->SolZenAng);

	return 0;
}



int write_fire_attributes(int32 sd_id, FireCounts *counts)
{
	int32 n_values = 1;

	if (SDsetattr(sd_id, "FirePix", DFNT_INT32, n_values, &counts->NumFirePix))
		return error(ERR_HDF);
	if (SDsetattr(sd_id, "MissingPix", DFNT_INT32, n_values, &counts->NumMissingPix))
		return error(ERR_HDF);
	if (SDsetattr(sd_id, "LandPix", DFNT_INT32, n_values, &counts->NumLandPix))
		return error(ERR_HDF);
	if (SDsetattr(sd_id, "WaterPix", DFNT_INT32, n_values, &counts->NumWaterPix))
		return error(ERR_HDF);

	if (SDsetattr(sd_id, "WaterAdjacentFirePix", DFNT_INT32, n_values, &counts->NumWaterAdjFirePix))
		return error(ERR_HDF);
	if (SDsetattr(sd_id, "CloudAdjacentFirePix", DFNT_INT32, n_values, &counts->NumCloudAdjFirePix))
		return error(ERR_HDF);

	if (SDsetattr(sd_id, "UnknownPix", DFNT_INT32, n_values, &counts->NumUnknownPix))
		return error(ERR_HDF);

	if (SDsetattr(sd_id, "LandCloudPix", DFNT_INT32, n_values, &counts->NumLandCloudPix))
		return error(ERR_HDF);
	if (SDsetattr(sd_id, "WaterCloudPix", DFNT_INT32, n_values, &counts->NumWaterCloudPix))
 		return error(ERR_HDF);
	if (SDsetattr(sd_id, "GlintPix", DFNT_INT32, n_values, &counts->NumGlintPix))
		return error(ERR_HDF);

	if (SDsetattr(sd_id, "GlintRejectedPix", DFNT_INT32, n_values, &counts->NumGlintRejectPix))
		return error(ERR_HDF);
	if (SDsetattr(sd_id, "CoastRejectedPix", DFNT_INT32, n_values, &counts->NumCoastRejectPix))
		return error(ERR_HDF);
	if (SDsetattr(sd_id, "HotSurfRejectedPix", DFNT_INT32, n_values, &counts->NumHotRejectPix))
		return error(ERR_HDF);


	if (SDsetattr(sd_id, "DayPix", DFNT_INT32, n_values, &counts->NumDayPix))
		return error(ERR_HDF);
	if (SDsetattr(sd_id, "NightPix", DFNT_INT32, n_values, &counts->NumNightPix))
		return error(ERR_HDF);

	return 0;
}



/* internal missing value of -999 degrees assigned to missing data pixels */
static void unscale_angle(int16 iang[LINES_PER_SCAN_1KM][SAMPLES_1KM],
	float ang[LINES_PER_SCAN_1KM][SAMPLES_1KM], float scale)
{
	int i, j;
	const int16 missing = -32767;

	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		for (j = 0; j < SAMPLES_1KM; j++)
			if (ang[i][j] != missing)
				ang[i][j] = scale * iang[i][j];
			else
				ang[i][j] = -999.0F;
		}
}




/* pass as 2D arrays so we can easily handle dead detectors */
/* can also be used to unscale reflectances merely by calling with reflectance
scales and offsets */
static void unscale(uint16 Lscaled[LINES_PER_SCAN_1KM][SAMPLES_1KM],
	float L[LINES_PER_SCAN_1KM][SAMPLES_1KM], float scale, float offset,
	float Lsat, UBYTE dead[LINES_PER_SCAN_1KM])
{
	int i, j;
	const uint16 SATURATED = 65533;

	/* this value can only occur in the aggregated bands */
	const uint16 AGG_FAILURE = 65528;

	const float Lmiss = -1.0F;

	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		if (dead[i]) {
			for (j = 0; j < SAMPLES_1KM; j++) L[i][j] = Lmiss;
			continue;
			}

		for (j = 0; j < SAMPLES_1KM; j++) {
			if ((Lscaled[i][j] > 32767) && (Lscaled[i][j] != SATURATED) &&
				(Lscaled[i][j] != AGG_FAILURE)) {
				L[i][j] = Lmiss;
				continue;
				}

			if ((Lscaled[i][j] != SATURATED) && (Lscaled[i][j] != AGG_FAILURE))
				L[i][j] = (Lscaled[i][j] - offset) * scale;
			else
				L[i][j] = Lsat;
			}
		}
}



int write_generic_attributes(int32 sd_id, char *infile_mod02, char *infile_mod03,
	enum Satellite sat)
{
	int32 n_values;
	char *s;

	s = satstr(sat);
	n_values = strlen(s);
	if (SDsetattr(sd_id, "Satellite", DFNT_CHAR8, n_values, s))
		return error(ERR_HDF);

	s = PROCESS_VERSION_NUMBER;
	n_values = strlen(s);
	if (SDsetattr(sd_id, "ProcessVersionNumber", DFNT_CHAR8, n_values, s))
		return error(ERR_HDF);

	n_values = strlen(infile_mod02);
	if (SDsetattr(sd_id, "MOD021KM input file", DFNT_CHAR8, n_values, infile_mod02))
		return error(ERR_HDF);

	n_values = strlen(infile_mod03);
	if (SDsetattr(sd_id, "MOD03 input file", DFNT_CHAR8, n_values, infile_mod03))
		return error(ERR_HDF);

	s = make_uname();
	n_values = strlen(s);
	if (SDsetattr(sd_id, "SystemID", DFNT_CHAR8, n_values, s))
		return error(ERR_HDF);

	return 0;
}


/*
Use MOD03/MYD03 metadata to find name of satellite (Terra or Aqua).

First try the simple way: look for a global attribute named
"ASSOCIATEDPLATFORMSHORTNAME", which is included in the Lynnes
"rationalized ECS" flavor of the 1A geolocation product.
If this attribute doesn't exist, we must (grudgingly) read and
parse the monstrous ECS "CoreMetadata.0" PVL string, which is also
stored as a global attribute.  If neither attribute is found (which
should never happen), or if the satellite name that was found is
not either "Terra" or "Aqua" (ignoring case), the satellite is
identified as SAT_UNKNOWN.

Returns 0 if no errors, otherwise returns non-zero error code.
*/
static int identify_satellite(int32 sd_id_mod03, enum Satellite *sat)
{
	int32 st;
	int32 attr_index;
	int32 data_type, n_values;
	char dummy[MAX_NC_NAME];
	char sat_name[8];
	char *coremeta;
	char *cptr;
	int i;
	const char attr_name[] = "ASSOCIATEDPLATFORMSHORTNAME";


	/* set this now to simplify error handling */
	*sat = SAT_UNKNOWN;


	attr_index = SDfindattr(sd_id_mod03, attr_name);
	if (attr_index != FAIL) {
		st = SDattrinfo(sd_id_mod03, attr_index, dummy,
			&data_type, &n_values);
		if (st == FAIL) return error(ERR_HDF);

		/* name can't possibly be "Terra" or "Aqua" if it is
		more than 5 characters in length */
		if (n_values >= 8) return 0;

		st = SDreadattr(sd_id_mod03, attr_index, sat_name);
		if (st == FAIL) return error(ERR_HDF);

		sat_name[n_values] = '\0';	/* add terminating null */
		}
	else {
		attr_index = SDfindattr(sd_id_mod03, "CoreMetadata.0");
		if (attr_index == FAIL) return 0;


		st = SDattrinfo(sd_id_mod03, attr_index, dummy,
			&data_type, &n_values);
		if (st == FAIL) return error(ERR_HDF);


		/* add extra character for terminating null */
		coremeta = (char *) malloc((n_values + 1) * sizeof(char));
		if (!coremeta) return error(ERR_F_MEM);

		st = SDreadattr(sd_id_mod03, attr_index, coremeta);
		if (st == FAIL) return error(ERR_HDF);

		coremeta[n_values] = '\0';

/*
Here's what we are looking for:

      OBJECT                 = ASSOCIATEDPLATFORMSHORTNAME
        CLASS                = "1"
        NUM_VAL              = 1
        VALUE                = "Aqua"
      END_OBJECT             = ASSOCIATEDPLATFORMSHORTNAME
*/

		cptr = strstr(coremeta, attr_name);
		if (!cptr) {
			free(coremeta);
			return 0;
			}

		cptr = strstr(cptr, " VALUE ");
		if (!cptr) {
			free(coremeta);
			return 0;
			}

		/* find opening quote */
		while (*cptr && (*cptr != '\"')) cptr++;

		if (!cptr) {
			free(coremeta);
			return 0;
			}

		/*
		Copy characters between quotation marks, up to a maximum
		of 7 characters.  If more than 7 characters, name will be
		truncated. */
		cptr++;
		i = 0;
		while (*cptr && (*cptr != '\"') && (i < 8))
			sat_name[i++] = *cptr++;

		sat_name[i] = '\0';

		free(coremeta);
		}


	if (!strcasecmp(sat_name, "Terra"))
		*sat = SAT_TERRA;
	else
		if (!strcasecmp(sat_name, "Aqua"))
			*sat = SAT_AQUA;

	return 0;
}




/* return pointer to static string naming satellite */
static char *satstr(enum Satellite sat)
{
	if (sat == SAT_TERRA) return "Terra";
	if (sat == SAT_AQUA) return "Aqua";

	return "?";
}


