/*
!Description:

	Implementation of alternate MODIS active fire detection
	algorithm described in forthcoming paper by Giglio, et al.

!References and Credits:

	L. Giglio and J. Descloitres, August 2001


!Design Notes:

	Although it would be better if it did not, this software was required
	by ECS to make use of the Hierarchical Data Format software created
	and distributed by the National Center for Supercomputing
	Applications.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>


#include "mfhdf.h"


#include "const.h"
#include "types.h"
#include "tb.h"
#include "fire.h"
#include "detfire.h"
#include "firetab.h"
#include "io.h"
#include "process.h"
#include "cmg.h"

#include "error.h"
#include "version.h"







int verbose;		/* flag: verbose mode */
int write_coarse;	/* flag: generate separate coarse-resolution product (MOD14CRS) */
int write_cmg;		/* flag: append CMG input-data layer to L2 product */
int write_tb;		/* flag: append brightness temperature SDSs to L2 product */
int dump_fires;		/* flag: */
int compress_sds;	/* flag: use internal HDF compression to compress 2D SDSs */


#ifdef OLD
#define EMISS_1KM_BANDS 16
#define REF_250M_BANDS 2

/* 1 km emissive bands radiance scales and offsets */
float radscale[EMISS_1KM_BANDS], radoff[EMISS_1KM_BANDS];

/* aggregated 250 m reflectance scales and offsets */
float refscale[REF_250M_BANDS], refoff[REF_250M_BANDS];
#endif




int main(int argc, char **argv)
{
	int32 sd_mod02, sd_mod03;
	int32 sd_fire;
	int32 sd_coarse_fire;	/* MOD14CRS */

	char *infile_mod02;	/* 1 km 1B product */
	char *infile_mod03;

	HDF_InputInfo hin;
	HDF_OutputInfo hout;

	char *outfile, *outfile_coarse;
	int st;

	char msg[MAX_MSG_LEN];

	FireCounts counts;

	int c;
	extern char *optarg;
	extern int optind, opterr;

#ifdef DEBUG
	int i;
#endif


	write_coarse = 0;
	outfile_coarse = (char *) NULL;
	verbose = write_tb = write_cmg = dump_fires = compress_sds = 0;

	while ((c = getopt(argc, argv, "vtc:dgC")) != -1) {
		switch (c) {
			case 'v':
				verbose = -1;
				break;

			case 't':
				write_tb = -1;
				break;

			case 'c':
				write_coarse = -1;
				outfile_coarse = optarg;
				break;

			case 'd':
				dump_fires = -1;
				break;

			case 'g':
				write_cmg = -1;
				break;

			case 'C':
				compress_sds = -1;
				break;

			default:
				USAGE:
				fputs("Usage: mod14 [-tvgdC] [-c coarse_output] MOD021KM_input"
					" MOD03_input output_file\n", stderr);
				fputs("    v    Verbose mode.\n", stderr);
				fputs("    t    Append T21,T31,T32 brightness temperature SDSs to L2 product.\n", stderr);
				fputs("    c    Write specified coarse-resolution product file.\n", stderr);
				fputs("    C    Enable internal HDF compression.\n", stderr);
				fputs("    g    Append CMG input-data SDS to L2 product.\n", stderr);
				fprintf(stderr, "\nVersion %s, compiled %s %s.\n",
					PROCESS_VERSION_NUMBER, __DATE__, __TIME__);
				exit(1);
				break;
			}
		}


	/* make sure three file names follow any switches */
	if (optind != argc - 3) goto USAGE;

	infile_mod02 = argv[optind];
	infile_mod03 = argv[optind+1];
	outfile = argv[optind+2];


	if (verbose) printf("%s\n%s\n%s\n", infile_mod02, infile_mod03, outfile);


	if ((sd_mod02 = SDstart(infile_mod02, DFACC_READ)) == FAIL) {
		(void) error(ERR_HDF_QUIET);
		(void) sprintf(msg, "Unable to open HDF input file \"%s\".", infile_mod02);
		errmsg(msg);
		exit(1);
		}

	if ((sd_mod03 = SDstart(infile_mod03, DFACC_READ)) == FAIL) {
		(void) error(ERR_HDF_QUIET);
		(void) sprintf(msg, "Unable to open HDF input file \"%s\".", infile_mod03);
		errmsg(msg);
		exit(2);
		}

	if ( (st = InitInputHDF(sd_mod02, sd_mod03, &hin)) ) exit(st);

	if (verbose) printf("nscans = %d\n", hin.nscans);

	if ((sd_fire = SDstart(outfile, DFACC_CREATE)) == FAIL) {
		(void) error(ERR_HDF_QUIET);
		(void) sprintf(msg, "Unable to open HDF output file \"%s\".", outfile);
		errmsg(msg);
		exit(3);
		}

	if (write_coarse) {
		if ((sd_coarse_fire = SDstart(outfile_coarse, DFACC_CREATE)) == FAIL) {
			(void) error(ERR_HDF_QUIET);
			(void) sprintf(msg, "Unable to open HDF output file \"%s\".", outfile);
			errmsg(msg);
			exit(3);
			}
		}
	else
		sd_coarse_fire = -1;

	st = InitOutputHDF(sd_fire, sd_coarse_fire, hin.nscans, &hout);
	if (st) exit(4);


	if (cmg_init(write_cmg)) {
		exit(4);
		}

	st = ProcGranule(&hin, &hout, &counts);
	if (st) {
		fputs("Error processing granule.\n", stderr);
		exit(3);
		}

	if (verbose) {
		printf("Fire:       %7d\n", counts.NumFirePix);
		printf("Unknown:    %7d\n", counts.NumUnknownPix);
		}

	if (write_cmg) {
if (verbose) puts("Writing CMG layers...");
		if (cmg_write(hout.sd_id_fire, CMG_DAY)) exit(5);
		if (cmg_write(hout.sd_id_fire, CMG_NIGHT)) exit(5);
		}

	if (write_generic_attributes(hout.sd_id_fire, infile_mod02,
		infile_mod03, hin.sat)) exit(5);


	if (write_coarse)
		if (SDend(sd_coarse_fire) == FAIL) exit(error(ERR_HDF));
	if (SDend(sd_fire) == FAIL) exit(error(ERR_HDF));
	if (SDend(sd_mod03) == FAIL) exit(error(ERR_HDF));
	if (SDend(sd_mod02) == FAIL) exit(error(ERR_HDF));

	free_fire_pixel_table();

	cmg_free(write_cmg);

	exit(0);
}


