/*
!C******************************************************************
!Description:

	Functions related to processing of each granule.

!Revision History:
!Team-unique Header:
!References and Credits:

	L. Giglio, Jan. 2001

!Design Notes:

	Although it would be better if it did not, this software was required
	by ECS to make use of the Hierarchical Data Format software created
	and distributed by the National Center for Supercomputing
	Applications.

!END*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>


#include "mfhdf.h"


#include "const.h"
#include "types.h"
#include "tb.h"
#include "fire.h"
#include "detfire.h"
#include "firetab.h"
#include "io.h"
#include "process.h"
#include "cmg.h"


#ifdef OLD_RED_HAT
#include "old_red_hat.h"
#endif

#ifdef NO_SINGLE_MATH
#include "no_sing_math.h"
#endif 


static void clear_scan(Scan *scan);
static void update_cmg(Scan *scan);


static Scan scanbuff[3];	/* triple buffer */



/* output buffer for the fire product */
extern uint8    FireMask[LINES_PER_SCAN_1KM][SAMPLES_1KM];


/* ---------- global flags ---------- */

extern int verbose;
extern int write_coarse;
extern int write_cmg;



int ProcGranule(HDF_InputInfo *hinfo, HDF_OutputInfo *hfire, FireCounts *counts)
{
	int iscan;
	int st;
	Scan *PrevPtr, *CurrPtr, *NextPtr;
	Scan *save;


#ifdef UNUSED
	Scan *scanptr[3];	/*  0 = previous, 1 = current, 2 = next */
#endif

	int first_line;
	static FireCounts ZeroFireCounts={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};


	PrevPtr = &scanbuff[0];
	CurrPtr = &scanbuff[1];
	NextPtr = &scanbuff[2];

	*counts = ZeroFireCounts;

	/* handle zero scans */
	if (hinfo->nscans < 1) {
		if ( (st = write_fire_attributes(hfire->sd_id_fire, counts)) )
			return st;

		if (write_coarse)
			if ( (st = write_fire_attributes(hfire->sd_id_coarse, counts)) )
				return st;

		return 0;
		}

	new_fire_pixel_table();


	/* clear previous buffer */
	clear_scan(PrevPtr);

	/* fill current buffer */
	st = ReadScan(hinfo, 0, CurrPtr);
	if (st) return st;


	/* screen current buffer */
	prescreen(CurrPtr);

	first_line = 0;
	for (iscan = 1; iscan < hinfo->nscans; iscan++) {
#ifdef DEBUG
		if (verbose) if (!(iscan % 10)) printf("%d\n", iscan);
#endif

		/* fill next buffer */
		st = ReadScan(hinfo, iscan, NextPtr);
		if (st) return st;


		/* screen next buffer */
		prescreen(NextPtr);

		/* do detection etc. on currptr */
		st = detect(PrevPtr, CurrPtr, NextPtr, first_line, counts);
		if (st) return st;

		if ( (st = WriteScan(hfire, iscan - 1, CurrPtr)) ) return st;

		if (write_cmg) {
			update_cmg(CurrPtr);
			}


		save = PrevPtr;
		PrevPtr = CurrPtr;
		CurrPtr = NextPtr;
		NextPtr = save;

		/*
		At this point scan buffer pointed to by NextPtr has crap
		in it that will be overwritten.
		*/

		first_line += LINES_PER_SCAN_1KM;
		}

	clear_scan(NextPtr);

	/* do detection on current buffer */
	st = detect(PrevPtr, CurrPtr, NextPtr, first_line, counts);
	if (st) return st;

	if ( (st = WriteScan(hfire, iscan - 1, CurrPtr)) ) return st;
if (write_cmg) {
	update_cmg(CurrPtr);
	}

	st = write_fire_pixel_table(hfire->sd_id_fire);
	if (st) return st;

	counts->NumPix = hinfo->nscans * LINES_PER_SCAN_1KM * SAMPLES_1KM;

	st = write_fire_attributes(hfire->sd_id_fire, counts);
	if (st) return st;

	if (write_coarse)
		if ( (st = write_fire_attributes(hfire->sd_id_coarse, counts)) )
			return st;
	return 0;
}



void compute_reflectances(float R1[LINES_PER_SCAN_1KM][SAMPLES_1KM],
	float R2[LINES_PER_SCAN_1KM][SAMPLES_1KM],
	float R7[LINES_PER_SCAN_1KM][SAMPLES_1KM],
	float SolZenAng[LINES_PER_SCAN_1KM][SAMPLES_1KM])
{
	int i, j;
	float mu_s;
	float inv_mu_s;


	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		for (j = 0; j < SAMPLES_1KM; j++) {
			/* handle nighttime observations */
			if (SolZenAng[i][j] > 85.0F) {
				R1[i][j] = R2[i][j] = -1.0F;
				continue;
				}

			mu_s = cosf(D2R * SolZenAng[i][j]);
			inv_mu_s = 1.0F / mu_s;

			if (R1[i][j] >= 0.0F)
				R1[i][j] *= inv_mu_s;

			if (R2[i][j] >= 0.0F)
				R2[i][j] *= inv_mu_s;

			if (R7[i][j] >= 0.0F)
				R7[i][j] *= inv_mu_s;
			}
		}
}




void compute_tb(float L[LINES_PER_SCAN_1KM][SAMPLES_1KM],
	float Tb[LINES_PER_SCAN_1KM][SAMPLES_1KM], float lambda)
/*
!C*****************************************************************************
!Description:
	Compute brightness temperatures for buffered radiances.

!Input Parameters:
	L	radiance (W m^-2 sr^-1 micrometer^-1)
	lambda	central wavelength (micrometers)

!Output Parameters:
	Tb	brightness temperature (K)

!Input/Output Parameters:
	None.

!Returns:
	Nothing.

!Revision History
	Original version February 2001, L. Giglio.

!Team-unique Header:

!References and Credits

!Design Notes

!END***************************************************************************
*/
{
	int i, j;

	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		for (j = 0; j < SAMPLES_1KM; j++) {
			if (L[i][j] > 0.0F)
				Tb[i][j] = InversePlanck(L[i][j], lambda);
			else
				Tb[i][j] = -1.0F;
			}
		}
}




void correct_tb(float Tb[LINES_PER_SCAN_1KM][SAMPLES_1KM], float *c)
/*
!C*****************************************************************************
!Description:
	Correct brightness temperatures computing via central wavelength
	for spectral response.

!Input Parameters:
	c	Correction coefficients; c[0] is slope, c[1] is intercept.

!Output Parameters:
	None.

!Input/Output Parameters:
	Tb	brightness temperature (K)

!Returns:
	Nothing.

!Revision History
	L. Giglio.

!Team-unique Header:

!References and Credits

!Design Notes

!END***************************************************************************
*/
{
	int i, j;

	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		for (j = 0; j < SAMPLES_1KM; j++) {
			if (Tb[i][j] > 0.0F)
				Tb[i][j] = Tb[i][j] * c[0] + c[1];
			}
		}
}



/* fill all fields of specified scan with missing data values */
static void clear_scan(Scan *scan)
/*
!C*****************************************************************************
!Description:
	Fill specified scan buffer with missing data values.

!Input Parameters:
	None.

!Output Parameters:
	scan	pointer to scan

!Input/Output Parameters:
	None.

!Returns:
	Nothing.

!Revision History
	Original version January 2001, L. Giglio.

!Team-unique Header:

!References and Credits

!Design Notes

!END***************************************************************************
*/
{
	int i, j;
	static PixInfo NullPixInfo = {0,0,0,0,0,0,0,0};

	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		for (j = 0; j < SAMPLES_1KM; j++) {
			scan->L21[i][j] = scan->L22[i][j] = -1.0F;
			scan->T21[i][j] = scan->T22[i][j] = -1.0F;
			scan->T31[i][j] = scan->T32[i][j] = -1.0F;
			scan->R1[i][j] = scan->R2[i][j] = -1.0F;
			scan->R7[i][j] = -1.0F;

			scan->lat[i][j] = scan->lon[i][j] = -999.9F;

			scan->SolZenAng[i][j] = scan->ViewZenAng[i][j]
				= scan->SolAzAng[i][j] = scan->ViewAzAng[i][j] = -999.0F;

			scan->landsea[i][j] = 0;

			scan->pixflags[i][j] = NullPixInfo;
			}
		}
}




static void update_cmg(Scan *scan)
{
	int i, j;

	for (i = 0; i < LINES_PER_SCAN_1KM; i++) {
		for (j = 0; j < SAMPLES_1KM; j++)
			cmg_append(scan->lat[i][j], scan->lon[i][j],
				FireMask[i][j], scan->pixflags[i][j].day);
		}
}

